import numpy as np
from copy import deepcopy
import fnmatch
import matplotlib.pyplot as plt
from itertools import groupby
import os
from textwrap import dedent
from .histogram import histogram1d as _histogram1d
from .histogram import histogram2d as _histogram2d
from .heatmap import _make_text, _map_to_equispaced
from matplotlib.colors import Normalize



def _make_1d_controlplot(histogram, nominal, reference, array, keys, var_hi, var_lo, path):
    """Helper function that saves a controlplot for a 1-dimensional histogram."""
    fig, ax = plt.subplots(1)
    x = histogram.points()[0]
    linestyles = ['-', '--', '-.', ':']
    if len(x) == 1:
        x = [0., 1.]
        for row_index in range(array.shape[0]):
            # There are 10 different colours in the default sequence, so change line style every 10 uncertainties to keep them distinguishable
            linestyle = linestyles[(row_index % 40) // 10] # index switches 0, 1, 2, 3, 0, 1, 2, 3, 0 etc. every 10 steps
            ax.plot(x, [array[row_index][0]/nominal[0], array[row_index][0]/nominal[0]], linestyle=linestyle, alpha=0.5, label=keys[row_index])
        ax.plot(x, [reference[0]/nominal[0], reference[0]/nominal[0]], 'r-.', label='Reference')
        ax.plot(x, [nominal[0]/nominal[0], nominal[0]/nominal[0]], 'k:', label='Nominal')
        ax.fill_between(x, [var_hi[0]/nominal[0], var_hi[0]/nominal[0]], [var_lo[0]/nominal[0], var_lo[0]/nominal[0]], facecolor='0.75', edgecolor='0.75', alpha=0.3, linewidth=0, label='Combined (incl. postprocessing)')
        ax.xaxis.set_ticklabels([])
        ax.set_xlabel('Integrated', ha='right', x=1., size='x-large')
    else:
        for row_index in range(array.shape[0]):
            # There are 10 different colours in the default sequence, so change line style every 10 uncertainties to keep them distinguishable
            linestyle = linestyles[(row_index % 40) // 10] # index switches 0, 1, 2, 3, 0, 1, 2, 3, 0 etc. every 10 steps
            ax.plot(x, array[row_index]/nominal, linestyle=linestyle, alpha=0.5, label=keys[row_index])
        ax.plot(x, reference/nominal, 'r-.', label='Reference')
        ax.plot(x, nominal/nominal, 'k:', label='Nominal')
        ax.fill_between(x, var_hi/nominal, var_lo/nominal, facecolor='0.75', edgecolor='0.75', alpha=0.3, linewidth=0, label='Combined (incl. postprocessing)')
        xlabel = histogram.attributes.get('xlabel', 'unknown')
        ax.set_xlabel(xlabel, ha='right', x=1., size='x-large')

    fig.subplots_adjust(right=0.5)
    ax.legend(bbox_to_anchor=(1.01, 1.15), fontsize='xx-small')
    ax.set_ylabel('Ratio to nominal', ha='right', y=1., size='x-large')
    # print(self.controlplot + ' {0}'.format(array.shape[0]))
    fig.savefig(path)
    plt.close(fig)



def _make_2d_controlplot(histogram, nominal, reference, array, keys, var_hi, var_lo, path):
    """Helper function that saves a controlplot for a 2-dimensional histogram.

    Caution: since there's one panel (= subplot) per variation in this plot, it will
    become very large if there are lots of variations.
    """
    nrows = int(np.ceil(len(keys)/2))
    fig, axes = plt.subplots(nrows, 2, figsize=(10, 4*nrows), squeeze=False, constrained_layout=True)
    # fig, axes = plt.subplots(len(keys), figsize=(8, 5*len(keys)))
    for index, (key, varied_areas) in enumerate(zip(keys, array)):

        these_axes = axes[np.unravel_index(index, axes.shape)]

        #print(key, varied_areas)
        uncertainty_hist = deepcopy(histogram)
        uncertainty_hist.corr_variations = {}
        uncertainty_hist.uncorr_variations = {}
        uncertainty_hist.areas = (varied_areas - nominal) / nominal * 100.0 # for percentage!
        #print(uncertainty_hist.areas)

        midx, midy, heights = uncertainty_hist.points()

        # Visualised bin edges and midpoints
        vis_binedges = uncertainty_hist.binedges
        vis_midx = midx
        vis_midy = midy
        # We use monowidth visualisation, where the actual bin edges and midpoints are mapped to equidistant
        # visualised ones:
        # Bin edges at integers from zero:
        vis_binedges = (np.array([i for i, _ in enumerate(uncertainty_hist.binedges[0])]), np.array([i for i, _ in enumerate(uncertainty_hist.binedges[1])]))
        # Bin midpoints at integers from zero PLUS 0.5 (to make them midpoints w.r.t. the edges):
        vis_midx = _map_to_equispaced(midx)
        vis_midy = _map_to_equispaced(midy)

        # midx and midy are used here as "dummy" fill values at the bin centres.
        # The histogram is filled "with" these values with the actual bin height (or optionally area) as "weight"
        nominals = np.ravel(uncertainty_hist.areas)

        these_axes.hist2d(vis_midx, vis_midy, weights=nominals, bins=vis_binedges, cmap='bwr', norm=Normalize(vmin=-2.0, vmax=2.0))

        # Write bin contents as text onto the plot
        def text_formatter(nominal=None, **ignore):
            text = f'{nominal:+.2f}'
            if 'nan' in text.lower():
                return ''
            return text
        for x, y, uncertainty_value in zip(vis_midx, vis_midy, nominals):
            text = _make_text(text_formatter, {'nominal' : uncertainty_value})
            these_axes.text(x, y, text, verticalalignment='center', horizontalalignment='center', color='black', fontsize=8)

        # Set proper tick labels for monowidth plots:
        these_axes.set_xticks(vis_binedges[0])
        these_axes.set_yticks(vis_binedges[1])
        # Nicely format the actual binedges: drop any trailing decimal points and zeros
        # if all have integer values by converting the values to ints:
        actual_binedges_x = uncertainty_hist.binedges[0].astype(int) if np.all(np.equal(np.mod(uncertainty_hist.binedges[0], 1), 0)) else np.around(uncertainty_hist.binedges[0], 5)
        actual_binedges_y = uncertainty_hist.binedges[1].astype(int) if np.all(np.equal(np.mod(uncertainty_hist.binedges[1], 1), 0)) else np.around(uncertainty_hist.binedges[1], 5)
        these_axes.set_xticklabels(actual_binedges_x)
        these_axes.set_yticklabels(actual_binedges_y)

        # Make the plot prettier
        these_axes.set_title(key, x=0., ha='left', size='large')
        if 'xlabel' in uncertainty_hist.plot_attributes:
            these_axes.set_xlabel(uncertainty_hist.plot_attributes['xlabel'], ha='right', x=1., size='large')
        if 'ylabel' in uncertainty_hist.plot_attributes:
            these_axes.set_ylabel(uncertainty_hist.plot_attributes['ylabel'], ha='right', y=1., size='large')
        these_axes.tick_params(which='both', bottom=True, top=True, left=True, right=True, direction='in')

    fig.suptitle('Relative uncertainties (%)', ha='center', size='x-large')
    # fig.subplots_adjust(left=0.08, bottom=0.08, right=0.08, top=0.08, wspace=0.08, hspace=0.08)
    fig.savefig(path)



# Terminology used in the combination methods:
# @array: Numpy array of varied histogram areas, where the columns correspond to bins and the rows to variations
# @nominal: Numpy array of the nominal histogram areas, with respect to which the uncertainty is determined

def combine_add_quad(array, nominal):
    diffs = array - nominal
    diffs_hi = diffs.clip(min=0.)
    diffs_lo = diffs.clip(max=0.)
    hi = nominal + np.sqrt(np.sum(diffs_hi**2, axis=0))
    lo = nominal - np.sqrt(np.sum(diffs_lo**2, axis=0))
    return (hi, lo)

def combine_add_lin(array, nominal):
    diffs = array - nominal
    diffs_hi = diffs.clip(min=0.)
    diffs_lo = diffs.clip(max=0.)
    hi = nominal + np.sum(diffs_hi, axis=0)
    lo = nominal + np.sum(diffs_lo, axis=0) # note: plus sign, the sum is non-positive
    return (hi, lo)

def combine_symm_rms(array, nominal):
    n_rows = array.shape[0] # how many variations (= rows) there are in the array
    shift = np.sqrt(np.sum((array - nominal)**2, axis=0) / float(n_rows))
    hi = nominal + shift
    lo = nominal - shift
    return (hi, lo)

def combine_asym_rms(array, nominal):
    n_hi = np.sum(np.greater_equal(array, nominal).astype(float), axis=0) # array containing the counts of high shifts in each bin
    shift_hi = np.sqrt(np.sum( ( (array - nominal).clip(min=0.) )**2, axis=0 ) / n_hi)
    hi = nominal + shift_hi
    n_lo = np.sum(np.less_equal(array, nominal).astype(float), axis=0) # array containing the counts of low shifts in each bin
    shift_lo = np.sqrt(np.sum( ( (array - nominal).clip(max=0.) )**2, axis=0 ) / n_lo)
    lo = nominal - shift_lo
    return (hi, lo)

def combine_envelope(array, nominal):
    # This reshaping seems to be done implicitly for 1D arrays, but must be done
    # explicitly for 2D arrays:
    nominal = deepcopy(nominal).reshape(1, *nominal.shape)
    array_incl_nominal = np.vstack((array, nominal)) # append nominal to array
    hi = np.max(array_incl_nominal, axis=0)
    lo = np.min(array_incl_nominal, axis=0)
    return (hi, lo)

def combine_asym_hessian(array, nominal):
    raise NotImplementedError
    return (hi, lo)

def combine_symm_hessian(array, nominal):
    shift = np.sqrt(np.sum((array - nominal)**2, axis=0))
    hi = nominal + shift
    lo = nominal - shift
    return (hi, lo)

def combine_asym_hessian_pairwise(array, nominal):
    diffs = array - nominal
    diffs_odd = diffs[0::2]
    diffs_even = diffs[1::2]
    if not diffs_odd.shape == diffs_even.shape:
        raise RuntimeError()
    hi = nominal + np.sqrt(np.sum( (np.maximum(diffs_odd, diffs_even).clip(min=0.) )**2, axis=0) )
    lo = nominal - np.sqrt(np.sum( (np.maximum(-diffs_odd, -diffs_even).clip(min=0.) )**2, axis=0) )
    return (hi, lo)

def combine_symm_hessian_pairwise(array, nominal):
    raise NotImplementedError
    return (hi, lo)



def _make_variation_array(variation_dictionary, keys):
    return np.array([variation_dictionary[key] for key in keys])



class model(object):
    '''
    Model for combining multiple variations into one uncertainty.

    Contains information of which variations to combine how and
    what to call the result.
    '''
    def __init__(self, name, keys, strategy, reference=None, postprocess=None, suffixes=('__hi', '__lo'), controlplot=None, matches_required=None):
        '''
        Let 'old' be the histogram whose variations will be combined and 'new' the resulting histogram.

        @name: of the uncertainty, will become the key in new.corr_variations
        @keys: list of names of the variations, should be in old.corr_variations.keys(), OR string containing Unix-style wildcards (*, ?, [...], [!...])
               that match any non-zero number of old.corr_variations.keys()
        @strategy:  how to primarily combine the variations into one uncertainty. Valid options are:
                    "no_comb" --- no combination, just write out the same variations as were found for the input
                    "drop" --- remove the variations from the histogram
                    "add_quad" --- add the differences between the variations and the nominal in quadrature
                    "add_lin" --- add the differences between the variations and the nominal linearly
                    "symm_rms" --- take the root mean square deviation of the variations from the nominal as the symmetric uncertainty. _If_ the nominal
                                   corresponds to the sample mean of the variations, the result corresponds to the standard deviation
                                   * example use case: combining NNPDF PDF variations into uncertainty
                    "asym_rms" --- take the root mean square difference from the nominal on each side of the nominal
                                   (smaller/larger) as the uncertainty
                                   * example use case: combining a set of toy variations into an asymmetric uncertainty
                    "asym_hessian" --- asymmetric Hessian uncertainty
                                       * example use case: ???
                    "symm_hessian" --- symmetric Hessian uncertainty
                                       * example use case: combining PDF4LHC15_30 PDF variations into uncertainty
                    "asym_hessian_pairwise" --- asymmetric Hessian uncertainty for cases in which pairwise variations are given. There MUST be an even
                                                number of variations (this is checked by the code) and they must be sorted so that variations 1 and 2,
                                                variations 3 and 4, etc., form pairs (this cannot be checked by the code, so it's entirely the user's
                                                responsibility).
                                                * example use case: combining CT14 PDF variations into uncertainty
                    "symm_hessian_pairwise" --- symmetric Hessian uncertainty for cases in which pairwise variations are given. There MUST be an even
                                                number of variations (this is checked by the code) and they must be sorted so that variations 1 and 2,
                                                variations 3 and 4, etc., form pairs (this cannot be checked by the code, so it's entirely the user's
                                                responsibility).
                                                * example use case: ???
                    "envelope" --- take the envelope of all variations AND the nominal as the uncertainty
                                   * example use case: combining QCD renormalisation and factorisation scale variations into uncertainty
        @reference: Some uncertainties require a reference histogram to calculate, e.g. the PDF uncertainty band of a PDF set should be calculated around the
                    nominal histogram of that PDF set, which is not the same as the nominal histogram if the PDF set in question is not the nominal PDF set.
                    @reference can be used to choose the correct reference histogram by setting it to the (string) key of the variation histogram that should
                    be used as reference.
                    If @reference=None, the nominal histogram will be used as reference.
        @postprocess:    options are:
                         None --- no further processing after the primary @strategy has been applied
                         "max" --- symmetrise around nominal by mirroring the larger absolute difference between variation and nominal in each bin
                         <float> --- if a float is given, the deviation of the variation from the nominal in each bin is multiplied by this number
                                     * example use case: convert 90% confidence interval uncertainties given for CT14 PDF set to 68% confidence interval
                                       uncertainties by setting postprocess=1./1.645
        @suffixes:  which suffixes to append for the resulting high and low variation (in that order)
        @controlplot:   if a path is given, a plot summarising the uncertainty combination may be created and saved under this name (if the controlplots argument
                        of model.apply() is not None). A text file describing what uncertainties were found and how they were combined is also created.
                        Useful for technical debugging as well as understanding physical effects.
        @matches_required:  can be set to an integer to require exactly that number of variations being found for the model. Otherwise, a RuntimeError is raised.

        NOTE: in the future, could add smoothings etc.
        '''
        super(model, self).__init__()
        self.name = name
        self.keys = keys
        self.strategy = strategy
        self.reference_key = reference
        self.postprocess = postprocess
        self.suffixes = suffixes
        self.controlplot = controlplot
        self.matches_required = matches_required
        self.combination_functions = {
            'no_comb' : None,
            'rename' : None,
            'drop' : None,
            'add_quad' : combine_add_quad,
            'add_lin' : combine_add_lin,
            'symm_rms' : combine_symm_rms,
            'asym_rms' : combine_asym_rms,
            'asym_hessian' : combine_asym_hessian,
            'symm_hessian' : combine_symm_hessian,
            'asym_hessian_pairwise' : combine_asym_hessian_pairwise,
            'symm_hessian_pairwise' : combine_symm_hessian_pairwise,
            'envelope' : combine_envelope,
        }
        if not strategy in self.combination_functions.keys():
            raise RuntimeError('Invalid uncertainty combination strategy "{0}" in model, please pick one of "{1}"'.format(strategy, ", ".join(self.combination_functions.keys())))



    def _find_keys(self, histogram, controlplot_location=None):
        all_keys = histogram.corr_variations.keys()
        if controlplot_location and self.controlplot:
            logfile_contents = dedent('''
            All available keys in histogram:

            {all_keys}

            Model will try to select keys matching:

            {keys}
            ''')
            logpath = os.path.join(controlplot_location, os.path.splitext(self.controlplot)[0]+'.txt')
            if not os.path.exists(controlplot_location):
                os.makedirs(controlplot_location)
            with open(logpath, 'w') as logfile:
                logfile.write(logfile_contents.format(all_keys=all_keys, keys=self.keys))
                # print(logfile_contents.format(all_keys=all_keys, keys=self.keys))
        # If self.keys is a list:
        if isinstance(self.keys, list):
            missing = [key for key in self.keys if not key in all_keys]
            if missing:
                raise RuntimeError('Uncertainty combination model "{0}" cannot be applied: the following input variations are missing in the input histogram: "{1}"'.format(self.name, '", "'.join(missing)))
            return self.keys
        # Else assume self.keys is a string, possibly containing Unix-style wildcards:
        keys = fnmatch.filter(list(all_keys), self.keys)
        if self.matches_required and len(keys) != self.matches_required:
            raise RuntimeError('Uncertainty combination model "{0}" cannot be applied: require {1} keys in histogram.corr_variations matching expression "{2}", but found {3}.'.format(self.name, self.matches_required, self.keys, len(keys)))
        return sorted(keys)



    def apply(self, histogram, controlplot_location=None):
        '''
        WARNING: SIDE EFFECTS - this method will change the @histogram.corr_variations dictionary.
        @controlplots: if a directory (end with '/') or prefix is given, control plots will be stored there for models
                       that have them enabled (model.controlplot != None)
        '''
        # print(histogram.corr_variations.keys())
        if self.controlplot is None:
            controlplot_location = None
        keys = self._find_keys(histogram, controlplot_location=controlplot_location)
        if not keys:
            raise RuntimeError('Found no variations for uncertainty combination model "{0}"'.format(self.name))
        if self.strategy == 'no_comb':
            return
        if self.strategy == 'drop':
            for key in keys:
                histogram.corr_variations.pop(key)
            return
        nominal = histogram.areas # alias for readability
        reference = nominal if not self.reference_key else histogram.corr_variations[self.reference_key]
        # Make array in which rows correspond to the variations:
        array = _make_variation_array(histogram.corr_variations, keys)
        # Combine original variations into new variations:
        combination_function = self.combination_functions[self.strategy]
        var_hi, var_lo = combination_function(array, reference) # NOTE: the histogram @reference is used as the @nominal in the combination function!!!
        # Apply postprocessing if desired:
        if self.postprocess:
            if isinstance(self.postprocess, float):
                var_hi *= reference + self.postprocess * (var_hi - reference)
                var_lo *= reference + self.postprocess * (var_lo - reference)
            elif self.postprocess == 'max':
                # Note: 'max' postprocessing is with respect to nominal, not reference
                shift = np.maximum(np.abs(var_hi - nominal), np.abs(var_lo - nominal))
                var_hi = nominal + shift
                var_lo = nominal - shift
        # Add new variations to the histogram:
        histogram.corr_variations[self.name + self.suffixes[0]] = var_hi
        histogram.corr_variations[self.name + self.suffixes[1]] = var_lo
        # Make controlplot if requested
        if controlplot_location:
            controlplot_args = [histogram, nominal, reference, array, keys, var_hi, var_lo, os.path.join(controlplot_location, self.controlplot)]
            if isinstance(histogram, _histogram1d):
                _make_1d_controlplot(*controlplot_args)
            elif isinstance(histogram, _histogram2d):
                _make_2d_controlplot(*controlplot_args)
        # Drop the original variations from the input histogram:
        for key in keys:
            histogram.corr_variations.pop(key)



def _keep_largest_shift(shift_array):
    '''
    Keep the largest shift in
    '''
    pass



def _delete_substrings(string, substrings):
    '''
    TODO: convert to a function called "_delete_suffixes" that only deletes
          substrings _at the end_.
    '''
    for ss in substrings:
        string = string.replace(ss, '')
    return string



# def _matches_any(string, body, suffixes):
#     '''
#     Return true if `string` is equal to `body` plus any of the `suffixes`.
#     If no `suffixes` are given, return False.
#     '''
#     if not suffixes:
#         return False
#     for suffix in suffixes:
#         if string == body + suffix:
#             return True
#     return False



def _iterator_over_variation_keys_grouped_by_label(keys, suffixes):
    '''
    Generator yielding the label (= variation key without suffix such as'_1up', '_1down', e.g. 'jes_1up' -> 'jes')
    and a list of the variation keys matching that label.
    '''
    keys = sorted(keys)
    for label, matching_keys in groupby(keys, lambda x : _delete_substrings(x, suffixes) ):
        yield label, list(matching_keys)



def _remove_nonmaximal_shifts(var, envelope, nominal):
    '''
    Return a modified version of `var` in which the deviations from the `nominal` have been set to
    zero if they are smaller than the corresponding values in `envelope` (which is a tuple of high, low!).

    CAUTION: HAS A BUG --- if two variations have the an identical shift (sign and magnitude), both are kept!!!
    '''
    raise NotImplementedError
    zero = np.zeros_like(var)
    shifts = var - nominal
    neg_shifts = np.where(shifts < 0, shifts, zero)
    pos_shifts = np.where(shifts > 0, shifts, zero)
    neg_envelope_shifts = np.where((envelope[1] - nominal) < 0, envelope[1] - nominal, zero)
    pos_envelope_shifts = np.where((envelope[0] - nominal) > 0, envelope[0] - nominal, zero)
    # Now calculate the shifts to be used in the result:
    # TODO: float comparison not numerically safe!!! Replace with something with safer!!!
    cleaned_neg = np.where(neg_shifts == neg_envelope_shifts, neg_shifts, zero)
    cleaned_pos = np.where(pos_shifts == pos_envelope_shifts, neg_shifts, zero)
    return nominal + cleaned_neg + cleaned_pos



def remove_same_sign_shifts(histogram, suffixes=['_1up', '_1down', '_up', '_down']):
    '''
    Drop smaller same-sign correlated variation shifts from the nominal for any group of variations
    whose names differ only by (any number of occurrences of) any of the strings in :code:`matches`
    '''
    raise NotImplementedError
    out = deepcopy(histogram)
    corr_variation_names = list(out.corr_variations.keys())
    envelope_by_label = {} # values will be tuples of (high, low)
    for label, matching_keys in _iterator_over_variation_keys_grouped_by_label(corr_variation_names, suffixes):
        envelope_by_label[label] = combine_envelope(_make_variation_array(out.corr_variations, matching_keys), out.areas)

    for key, var in out.corr_variations.items():
        label = _delete_substrings(key, suffixes)
        out.corr_variations[key] = _remove_nonmaximal_shifts(var, envelope_by_label[label], out.areas)

    return out



def combine_copy(histogram, models, ignore_missing=False, controlplot_location=None, drop_same_sign_shifts=False, suffixes=['_1up', '_1down', '_up', '_down']):
    '''
    @histogram: the input histogram. The return value will be a copy of this histogram, with the desired variation combinations applied.
    @models: iterable of the models to be applied
    @ignore_missing: if True, do not throw an exception if the input variations specified in a model are missing, but simply ignore the model
    @controlplot_location: if a directory (end with '/') or prefix is given, control plots will be stored there for models
                           that have them enabled (model.controlplot != None)
    :param drop_same_sign_shifts: if True, _correlated_ variation names that differ only by any suffix given in argument :code:`suffixes` are grouped together. If more than one of these grouped variations has a shift with respect to the nominal in a given bin, only the largest shift in that bin is kept. The other shifts are set to zero (i.e. the variation is set to equal the nominal in the bin). A common case where this is useful is to avoid double-counting the same source of systematic uncertainty in bins where the "up" and "down" variation point in the same direction with respect to the nominal in some bin(s). Uncorrelated variations are not affected by this option.
    :type drop_same_sign_shifts: :code:`bool`
    :param suffixes: see argument :code:`drop_same_sign_shifts` for explanation
    :type suffixes: :code:`list` of :code:`str`

    IMPORTANT NOTE: it is possible to apply combination models whose input variations are only produced in the same call of combine_copy. In other words,
                    the variations don't yet need to exist in the @histogram.corr_variations when passing @histogram to combine_copy.
    '''
    if drop_same_sign_shifts:
        out = remove_same_sign_shifts(histogram, suffixes=suffixes)
    else:
        out = deepcopy(histogram)
    # print(list(out.corr_variations.keys()))

    for model in models:
        try:
            model.apply(out, controlplot_location=controlplot_location)
            # print(model.name)
            # print(list(out.corr_variations.keys()))
        except RuntimeError:
            if not ignore_missing:
                raise
    return out



import numpy as np
import numbers
import operator
import warnings
from copy import deepcopy
from functools import partial
from textwrap import dedent
import pandas as pd

from . import Value



def _calculate_bin_sizes(binedges):
    '''
    Auxiliary function for calculating bin sizes (widths in the 1D case, surface areas in the 2D case).

    :param binedges: bin edges
    :type binedges: Numpy array, tuple of two Numpy arrays
    :returns: Numpy array of bin sizes

    For a one-dimensional histogram, returns an array of dimension (N, 1), where N is the number of bins.
    The elements represent the width of each bin.

    For a two-dimensional histogram, returns an array of dimension (N, M), where N is the number of bins
    along the first axis ("x-axis") and M is the number of bins along the second axis ("y-axis"). The
    elements represent the surface area of each bin.

    '''
    if isinstance(binedges, tuple):
        if len(binedges) > 2:
            raise NotImplementedError('Calculating bin areas is only implemented for 1- and 2-dimensional histograms')
        # Calculate bin widths along each axis:
        binwidths = (np.diff(edges) for edges in binedges)
        return np.outer(*binwidths)
    return np.diff(binedges)



def _get_bin_index(binedges, value, label='x'):
    """Returns the index of the bin that contains the given value.

    Lower bin edges are included in a bin, upper bin edges are excluded (same as in the `ROOT <https://root.cern>`_ convention).

    :param binedges: the bin edges
    :type binegdes: ``np.array`` of ``float``
    :param value: value
    :type value: ``float``
    :param label: a label that makes error messages, if any, more useful. Typically the axis that the binedges and value correspond to.
    :type label: ``str``

    :returns: index of bin that contains the value
    :rtype: ``int``

    :raises ValueError: if value lies outside of the outer bin edges
    """
    if value < np.min(binedges) or value >= np.max(binedges):
        raise ValueError('Cannot find index of bin containing {label} = {val}, which is outside of histogram {label}-boundaries [{valmin}, {valmax})'.format(val=value, label=label, valmin=np.min(binedges), valmax=np.max(binedges)))
    return np.searchsorted(binedges, value, side='right') - 1





class basehistogram(object):
    '''
    Base class for one-dimensional and two-dimensional histograms that keep track of their various uncertainty contributions and arbitrary attributes (useful for labeling and plotting).

    :param binedges: bin edges, including uppermost. For 1D histograms, a :code:`numpy.array`. For 2D histograms, a :code:`tuple` of two :code:`numpy.array` (in the x and y direction, respectively).
    :type binedges: :code:`numpy.array`, or :code:`tuple` of :code:`numpy.array`
    :param contents: the "bin contents", which are either bin areas (= what ROOT calls "bin contents") or bin heights (= bin areas / bin sizes). See also argument :code:`areas`.
    :type contents: :code:`numpy.array`
    :param areas: if True, interpret given contents as bin areas, else as bin heights
    :type areas: :code:`bool`
    :param name: a name for the histogram. This is only separate from the other attributes because it is so commonly used and is automatically created for histograms produced by mathematically combining two histograms. E.g. dividing two histograms with names :code:`'foo'` and :code:`'bar'` will return a histogram with name :code:`'foo / bar'`.
    :type name: :code:`str`
    :param uncorr_variations: dictionary of variations that are uncorrelated between bins (e.g. statistical uncertainty). Keys are variation names, values are :code:`np.array` objects of the same dimension as the nominal :code:`contents`.
    :type uncorr_variations: :code:`dict`
    :param corr_variations: dictionary of variations that are fully correlated between bins (e.g. systematic uncertainty). Keys are variation names, values are :code:`np.array` objects of the same dimension as the nominal :code:`contents`.
    :type corr_variations: :code:`dict`
    :param attributes: dictionary of completely arbitrary attributes that the user can provide/change/access. E.g. information about the data sample that produced the histogram.
    :type attributes: :code:`dict`
    :param plot_attributes: dictionary of completely arbitrary that the user can provide/change/access. This one is more intended for information on how to visualise/plot the histogram. It is especially useful if working with `heppy.make_figure`, which will *assume* that all the plot_attributes correspond to keyword arguments that are understood by Matplotlib's :code:`plot()` and/or :code:`fill_between()` functions
    :type plot_attributes: :code:`dict`
    '''
    def __init__(self, binedges, contents, areas=False, name='', uncorr_variations={}, corr_variations={}, attributes={}, plot_attributes={}):
        super(basehistogram, self).__init__()
        self.binedges = deepcopy(binedges)
        binsizes = _calculate_bin_sizes(binedges)
        self.areas = np.array(deepcopy(contents)) if areas else np.array(deepcopy(contents)) * binsizes
        self.name = deepcopy(name)
        self.uncorr_variations = deepcopy(uncorr_variations) if areas else {key : deepcopy(heights) * binsizes for key, heights in uncorr_variations.items()}
        self.corr_variations = deepcopy(corr_variations) if areas else {key : deepcopy(heights) * binsizes for key, heights in corr_variations.items()}
        self.attributes = deepcopy(attributes)
        self.plot_attributes = deepcopy(plot_attributes)



    def __str__(self):
        '''
        Human-readable string showing the contents and properties of the histogram.
        '''
        properties = {
            'name' : self.name,
            'type' : type(self),
            'attributes' : '\n  '.join([key+' : '+val.__str__() for key, val in self.attributes.items()]),
            'plot_attributes' : '\n  '.join([key+' : '+val.__str__() for key, val in self.plot_attributes.items()]),
            'binedges' : self.binedges.__str__(),
            'binsizes' : self.binsizes.__str__(),
            'areas' : self.areas.__str__(),
            'uncorr_variations' : '\n  '.join([key+' : '+val.__str__() for key, val in self.uncorr_variations.items()]),
            'corr_variations' : '\n  '.join([key+' : '+val.__str__() for key, val in self.corr_variations.items()]),
            }
        for key in properties.keys():
            if properties[key] == '': properties[key] = '(none)'
        form = '\n'.join([
            'Name: {name}',
            'Type: {type}',
            'Attributes:',
            '  {attributes}',
            'Plot attributes:',
            '  {plot_attributes}',
            'Bin edges:',
            '  {binedges}',
            'Bin sizes:',
            '  {binsizes}',
            'Nominal areas:',
            '  {areas}',
            'Uncorrelated variation areas:',
            '  {uncorr_variations}',
            'Correlated variation areas:',
            '  {corr_variations}',
            ])
        return form.format(**properties)



    def extract_variation_histogram(self, variation, **kwargs):
        '''
        Get a new histogram object that has a given variation as nominal.
        Useful e.g. for studying a particular systematic variation.

        :param variation: name of the variation
        :type variation: :code:`str`
        :param **kwargs: get passed on to constructor of new histogram, e.g. useful to set a :code:`name` for the new histogram.

        :returns: new :py:class:`heppy.histogram` that has a given :code:`variation` as nominal
        :raises KeyError: if variation not found in either uncorrelated or correlated variations
        :raises RuntimeError: if variation found in both uncorrelated or correlated variations
        '''
        if variation in self.uncorr_variations.keys() and variation in self.corr_variations.keys():
            raise RuntimeError('Cannot resolve ambiguity: a variation of name "{variation}" was found in both the uncorrelated and correlated variation dictionary.'.format(variation=variation))
        try:
            return type(self)(self.binedges, self.corr_variations[variation], areas=True, **kwargs)
        except KeyError:
            try:
                return type(self)(self.binedges, self.uncorr_variations[variation], areas=True, **kwargs)
            except KeyError:
                raise KeyError('Variation "{variation}" was not found in either the uncorrelated or correlated variation dictionary.'.format(variation=variation))



    @property
    def binsizes(self):
        '''
        Bin sizes.

        For a one-dimensional histogram, returns an array of dimension (N, 1), where N is the number of bins.
        The elements represent the width of each bin.

        For a two-dimensional histogram, returns an array of dimension (N, M), where N is the number of bins
        along the first axis ("x-axis") and M is the number of bins along the second axis ("y-axis"). The
        elements represent the area of each bin.
        '''
        # For 2D histograms, self.binedges is a tuple of arrays that describe the bin edges along each axis
        return _calculate_bin_sizes(self.binedges)


    @property
    def heights(self):
        '''
        Bin heights, equal to bin areas divided by the corresponding bin sizes
        '''
        return self.areas / self.binsizes



    def set_heights(self, heights):
        '''
        Set bin heights to an array of the same dimension as the current areas or to a scalar
        '''
        assert heights.shape == self.binsizes.shape or isinstance(heights, numbers.Number)
        self.areas = self.binsizes * heights



    def integral(self, variations=None, **kwargs):
        '''
        Calculate the integral of the histogram.

        :param variations: if given, a tuple of the nominal integral and its upper and lower variation is calculated. \
        This argument is passed to :py:func:`histogram1d.net_variations` and should be a list of considered variation \
        names or the string :code:`'all'`.
        :type variations: :code:`list` of :code:`str` or :code:`str`
        :param **kwargs: additional keyword arguments that get passed to :py:func:`histogram1d.net_variations`

        :returns: the integral (as well as upper and lower variation if :code:`variations` is given)
        :rtype: :code:`float`, or if :code:`variations` are given, :code:`tuple` of nominal as well as upper and lower variation
        '''
        if not variations:
            return np.sum(self.areas)
        one_bin_histogram = deepcopy(self)
        one_bin_histogram.rebin(np.array([self.binedges[0], self.binedges[-1]]))
        up, down = one_bin_histogram.net_variations(variations=variations, **kwargs)
        return self.integral(), up[0], down[0]



    def net_variations(self, variations='all', subtract_nominal=False, relative=False):
        '''
        Return upper and lower net areas variation of the histogram as a tuple.

        @variations should be a sequence of considered variation names or the string 'all'
        @subtract_nominal: if True, return the differences with respect to the nominal areas
        @relative: if True, divide by the nominal areas

        CAUTION: this method cannot yet deal with systematic uncertainties for which the up- and down-shift lie on the same side of the nominal.
        This is because the variations are fundamentally treated independently of each other, so there is no sense of the up- and down-shift
        being related to the same underlying uncertainty source.
        '''
        all_variations = {**self.uncorr_variations, **self.corr_variations} # correlated/uncorrelated only matters for rebinning so far...
        if variations == 'all':
            variations = all_variations.keys()
        if not variations:
            return (np.zeros_like(self.areas), np.zeros_like(self.areas))
        v = np.array([all_variations[name] for name in all_variations if name in variations])
        #print('variations array')
        #print(v)
        shifts = v - self.areas
        lower = self.areas - np.sqrt(np.sum(shifts.clip(max=0)**2, axis=0)) # add negative shifts from nominal in quadrature
        #print('lower')
        #print(lower)
        upper = self.areas + np.sqrt(np.sum(shifts.clip(min=0)**2, axis=0)) # add positive shifts from nominal in quadrature
        #print('upper')
        #print(upper)
        if subtract_nominal:
            lower = lower - self.areas
            upper = upper - self.areas
        if relative:
            # If height == 0, division would give NaN. Using nan_to_num, this is replaced by zero.
            lower = np.nan_to_num(lower / self.areas)
            upper = np.nan_to_num(upper / self.areas)
        return upper, lower



    def errorbars(self, variations='all'):
        '''
        Returns upper and lower error bars, defined as the absolute net variations (taking into account
        the given variations) with the nominal values subtracted.
        '''
        upper, lower = self.net_variations(variations, subtract_nominal=True, relative=False)
        return np.abs(upper)/self.binsizes, np.abs(lower)/self.binsizes



    # Mathematical operations for combining multiple histograms are defined in the following



    def _infer_shift_directions(self, self_shifts, other_shifts):
        '''
        Auxiliary function that infers the direction of shifts when combining them for two histograms

        Robust under some shifts being zero (and hence having no direction, np.sign(0) = 0). This led
        to a bug in the past!
        '''
        self_signs = np.sign(self_shifts)
        other_signs = np.sign(other_shifts)
        with warnings.catch_warnings():
            warnings.simplefilter('ignore', RuntimeWarning) # do not warn about NaNs in the division in the following line
            return np.nan_to_num((self_signs + other_signs) / (np.abs(self_signs) + np.abs(other_signs)))



    def _modify_areas(self, other, operation, symbol):
        '''
        Backend function that internally implements the mathematical operations between two histograms
        Alternatively, @other may be a scalar (float or int)

        NOTE: if @operation is truediv, the bin heights rather than areas are set.

        Behaviour for combining UNCORRELATED uncertainties:

        '''
        scalar = False
        if isinstance(other, (int, float)):
            scalar = True
            other = type(self)(self.binedges, np.zeros_like(self.areas) + other, areas=True, name='{0}'.format(other))
        # NB: array_equal also works for a tuple of Numpy arrays, as we have for a 2D histogram!
        if isinstance(self.binedges, np.ndarray):
            if not np.allclose(self.binedges, other.binedges):
                raise RuntimeError('Cannot perform mathematical operation on histograms, binegdes differ:\n{a}\n{b}'.format(a=self.binedges, b=other.binedges))
        else:
            if not np.allclose(self.binedges[0], other.binedges[0]) or not np.allclose(self.binedges[1], other.binedges[1]):
                raise RuntimeError('Cannot perform mathematical operation on histograms, binegdes differ:\n{a}\n{b}'.format(a=self.binedges, b=other.binedges))
        selfname = '({0})'.format(self.name) if ' ' in self.name else self.name
        othername = '({0})'.format(other.name) if ' ' in other.name else other.name
        resultname = '{0} {1} {2}'.format(selfname, symbol, othername)
        attributes = {**other.attributes, **self.attributes}
        plot_attributes = {**other.plot_attributes, **self.plot_attributes}
        binedges = self.binedges
        with warnings.catch_warnings():
            warnings.simplefilter('ignore', RuntimeWarning) # do not warn about NaNs in the division in the following line
            areas = operation(self.areas, other.areas)
        # Constructing uncorrelated variations of the new histogram
        uncorr_variations = {}
        if scalar:
            # for name, array in self.uncorr_variations.items():
            #     print(name, array, other.areas, operation(array, other.areas))
            uncorr_variations = {name : operation(array, other.areas) for name, array in self.uncorr_variations.items()}
        elif not operation in [operator.add, operator.sub, operator.truediv]:
            # CAUTION: calculation of uncorrelated uncertainty is not implemented in this case
            pass
        else:
            # TODO: factor this else clause into a separate method!
            all_uncorr_variations = list(set(list(self.uncorr_variations.keys()) + list(other.uncorr_variations.keys())))
            for name in all_uncorr_variations:
                try:
                    self_shifts = self.uncorr_variations[name] - self.areas
                except KeyError:
                    self_shifts = np.zeros_like(self.areas)
                try:
                    other_shifts = other.uncorr_variations[name] - other.areas
                except KeyError:
                    other_shifts = np.zeros_like(other.areas)
                # CAUTION: the direction information (i.e. + or -) of the shifts is lost when summing in quadrature / dividing,
                # so it is ASSUMED that it can be taken from the direction of self_shifts. This is only true if all
                # the direction of a given shift for a given variation is the same for self and other.
                shift_directions = self._infer_shift_directions(self_shifts, other_shifts)
                if operation == operator.truediv:
                    with warnings.catch_warnings():
                        warnings.simplefilter('ignore', RuntimeWarning) # do not warn about NaNs in the division in the following line
                        uncorr_variations[name] = areas + shift_directions * np.sqrt((self_shifts / other.areas)**2 + (self.areas * other_shifts / other.areas**2)**2)
                else:
                    uncorr_variations[name] = areas + shift_directions * np.sqrt((self_shifts**2 + other_shifts**2))


        # Constructing correlated variations of the new histogram
        corr_variations = {}
        if scalar:
            corr_variations = {name : operation(array, other.areas) for name, array in self.corr_variations.items()}
        else:
            # TODO: factor this else clause into a separate method!
            all_corr_variations = list(set(list(self.corr_variations.keys()) + list(other.corr_variations.keys())))
            for name in all_corr_variations:
                try:
                    self_var = self.corr_variations[name]
                except KeyError:
                    # this variation was not found for histogram @self, use its nominals instead
                    self_var = self.areas
                try:
                    other_var = other.corr_variations[name]
                except KeyError:
                    # this variation was not found for histogram @other, use its nominals instead
                    other_var = other.areas
                # CAUTION: the direction information (i.e. + or -) of the shifts is lost when summing in quadrature,
                # so it is ASSUMED that it can be taken from the direction of self_shifts. This is only true if all
                # the direction of a given shift for a given variation is the same for self and other.
                corr_variations[name] = operation(self_var, other_var)

        #print('TODO: add support for combining variations in histogram manipulation')
        return type(self)(binedges, areas, name=resultname, areas=True, uncorr_variations=uncorr_variations, corr_variations=corr_variations, attributes=attributes, plot_attributes=plot_attributes)



    def __add__(self, other):
        '''
        Add another histogram or a scalar to this histogram.

        Returns the result of the addition as a histogram.

        Correlated variations are treated as fully correlated among the two histograms if they have the same
        name, otherwise they are treated as uncorrelated. Uncorrelated variations are treated as uncorrelated
        between the two histograms.
        '''
        return self._modify_areas(other, operator.add, '+')

    def __sub__(self, other):
        '''
        Subtract another histogram or a scalar from this histogram.

        Returns the result of the subtraction as a histogram.

        Correlated variations are treated as fully correlated among the two histograms if they have the same
        name, otherwise they are treated as uncorrelated. Uncorrelated variations are treated as uncorrelated
        between the two histograms.
        '''
        return self._modify_areas(other, operator.sub, '-')

    def __mul__(self, other):
        '''
        Multiply another histogram or a scalar binwise with this histogram.

        Returns the result of the binwise multiplication as a histogram.

        Correlated variations are treated as fully correlated among the two histograms if they have the same
        name, otherwise they are treated as uncorrelated. Uncorrelated variations are treated as uncorrelated
        between the two histograms.
        '''
        return self._modify_areas(other, operator.mul, '*')

    def __truediv__(self, other):
        '''
        Divide by another histogram or a scalar binwise.

        Returns the result of the binwise division as a histogram.

        Correlated variations are treated as fully correlated among the two histograms if they have the same
        name, otherwise they are treated as uncorrelated.

        CAUTION: Uncorrelated variations are treated as uncorrelated between the two histograms. If the
        uncorrelated variations represent statistical uncertainties, this means that the division treats
        the two histograms as statistically uncorrelated.

        See also
        :py:func:`histdiv`
        '''
        return self._modify_areas(other, operator.truediv, '/')



    def apply_inplace(self, function, new_binedges=None):
        r'''
        Call a function on the nominal areas as well as all varied areas (in :code:`corr_variations` and :code:`uncorr_variations`),
        modifying the existing histogram.

        It is possible to convert the histogram to a different type (e.g. :py:class:`histogram2d` :math:`\to` :py:class:`histogram1d`)
        by giving new binedges of the desired new dimensionality. If the new binedges have a dimension other than 1D or 2D,
        the type will become :py:class:`basehistogram`.

        For a version of this method that leaves the original histogram unchanged and returns a copy with the function applied (and
        optionally new bin edges), see :py:func:`basehistogram.apply`. I HIGHLY recommend using that if the new histogram will be of
        different type to avoid confusion!

        Caution: if you change the binning, it is your responsibility that uncertainties encoded in the variations are handled correctly.

        :param function: function taking a :code:`numpy.array` as argument
        :type function: :code:`function`
        :param new_binedges: optional argument to set new binedges. If the binedges change the dimension of the histogram (e.g. from 2D to 1D), the histograms' type is changed accordingly
        :type function: :code:`function`

        Example 1: taking the sine of the areas

        .. code-block:: python

            import heppy as hp
            import numpy as np
            foo = hp.histogram1d(np.array([0., 1., 2., 3.]), [5., 6., 7])
            foo.apply_inplace(np.sin)

        Example 2: projecting a 2D histogram to its x-axis (integrating over the y-axis)

        .. code-block:: python

            import heppy as hp
            import numpy as np
            binedges = (np.array([0., 10., 20.]), np.array([10., 20., 30.]))
            areas = [[0.1, 0.2], [0.3, 0.4]]
            foo = hp.histogram2d(binedges, areas, areas=True)
            from functools import partial
            project_x = partial(np.sum, axis=1)
            foo.apply_inplace(project_x, new_binedges=foo.binedges[0])

        :raises: :code:`ValueError` if the shape of the areas after the function is called on them does not match the shape of the bin edges (after setting them to :code:`new_binedges` if given)
        '''
        # Check that the shape of the areas matches the shape of the bin sizes (possibly after changing binning)
        if function(self.areas).shape != (self.binsizes.shape if new_binedges is None else _calculate_bin_sizes(new_binedges).shape):
            raise ValueError('Binedges do not match after applying function to areas')
        change_type = (new_binedges is not None and _calculate_bin_sizes(new_binedges).shape != self.binsizes.shape)
        if new_binedges is not None:
            self.binedges = deepcopy(new_binedges)
        self.areas = function(self.areas)
        for key, areas in self.corr_variations.items():
            self.corr_variations[key] = function(areas)
        for key, areas in self.uncorr_variations.items():
            self.uncorr_variations[key] = function(areas)
        # Cast the histogram to the right type based on what its bin dimensionality is now:
        if change_type:
            args = [self.binedges, self.heights]
            kwargs = {
                'name' : self.name,
                'uncorr_variations' : self.uncorr_variations,
                'corr_variations' : self.corr_variations,
                'attributes' : self.attributes,
                'plot_attributes' : self.plot_attributes,
            }
            if isinstance(new_binedges, np.ndarray):
                self.__class__ = histogram1d
            if isinstance(new_binedges, tuple) and len(new_binedges) == 2:
                self.__class__ = histogram2d
            else:
                self.__class__ = basehistogram



    def apply(self, function, new_binedges=None):
        '''
        Same as :py:func:`basehistogram.apply_inplace`, except that the resulting histogram is returned (as an independent object) with
        the function applied, while the original histogram is not modified.

        :returns: histogram with function applied and possibly new bin edges (:py:class:`histogram1d`, :py:class:`histogram2d`, or :py:class:`basehistogram`)
        '''
        new = deepcopy(self)
        new.apply_inplace(function, new_binedges=new_binedges)
        return new



    def clip(self, minimum=None, maximum=None):
        """Clips the bin areas at a minimum and/or maximum value.

        Similar to ``numpy.clip`` (see `here <https://numpy.org/doc/stable/reference/generated/numpy.clip.html>`_).

        Both the nominals and all correlated and uncorrelated variations of the histogram are clipped.
        *Note that the net variation of multiple individual variations considered together may still lie
        outside of the clip range!*

        :param minimum: minimum accepted value; areas below this value will be set to the value.
        If ``None``, no minimum is imposed.
        :type minimum: ``float`` or ``None``
        :param maximum: maximum accepted value; areas above this value will be set to the value.
        If ``None``, no maximum is imposed.
        :type maximum: ``float`` or ``None``

        The behaviour is like that of ``numpy.clip``, e.g. with regards to what happens if ``minimum`` is
        greater than ``maximum``.

        **"Undocumented" feature:** you can actually pass a Numpy ``array_like`` as ``minimum`` or ``maximum``,
        not just a single ``float``!
        """
        def clip_function(array):
            return np.clip(array, minimum, maximum)
        self.apply_inplace(clip_function)



    def __repr__(self):
        '''
        String representation of the histogram that can be used to recreate it, e.g. with the following code:

        .. code-block:: python

            from heppy import histogram1d
            from numpy import array
            foo = histogram1d(array([0., 1., 2., 3.]), array([5., 6., 7.])) # constructed with arbitrary options
            bar = eval(repr(foo)) # bar is equivalent to foo

        :returns: :code:`str` representation of the object
        '''
        histogram_type = 'histogram1d' if 'histogram1d' in str(type(self)) else 'histogram2d'
        return histogram_type + '(' + ', '.join([
            repr(self.binedges),
            repr(self.areas),
            'areas=True',
            'name={}'.format(repr(self.name)),
            'attributes={}'.format(repr(self.attributes)),
            'plot_attributes={}'.format(repr(self.plot_attributes)),
            'uncorr_variations={}'.format(repr(self.uncorr_variations)),
            'corr_variations={}'.format(repr(self.corr_variations)),
            ]).replace('\n', ' ') + ')'



    def to_file(self, outfilename, key, recreate=False):
        '''
        Write histogram to text file. Multiple histograms with different keys
        can be written to the same file.

        :param outfilename: name of the file that the histogram is written to
        :type outfilename: :code:`str`
        :param key: name/key of the histogram inside the output file
        :type key: :code:`str`
        :param recreate: option to recreate the output file rather than append to it
        :type key: :code:`bool`
        '''
        mode = 'w' if recreate else 'a'
        with open(outfilename, mode) as outfile:
            outfile.write(key + ' : ' + repr(self) + '\n')





class histogram1d(basehistogram):
    '''
    Heppy one-dimensional histogram.
    This has functionality for rebinning, getting various representations for plotting (curve, points, errorbars, errorbands), as well as
    performing mathematical operations (these have only been implemented for one-dimensional histograms so far).
    '''
    def __init__(self, *args, **kwargs):
        super(histogram, self).__init__(*args, **kwargs)



    @property
    def nbins(self):
        """Returns the number of bins in the histogram.

        :returns: number of bins in the histogram
        :rtype: ``int``
        """
        return len(self.binedges) - 1



    @property
    def binwidths(self):
        '''
        Bin widths is an alias for bin sizes in the case of a one-dimensional histogram
        '''
        return super(histogram1d, self).binsizes



    def bin_index(self, x):
        """Returns the index of the bin that contains the given x-value.

        Lower bin edges are included in a bin, upper bin edges are excluded (same as in the `ROOT <https://root.cern>`_ convention).

        :param x: x-value
        :type x: ``float``

        :returns: index of bin that contains the x-value
        :rtype: ``int``

        :raises ValueError: if x-value lies outside of the outer bin edges of the histogram

        Example:

        .. code-block:: python

            >>> h = histogram1d([0., 1., 2.], [10., 11.])
            >>> h.bin_index(0.5)
            0
            >>> h.bin_index(0.)
            0
            >>> h.bin_index(1.0)
            1
            >>> h.bin_index(2.0)
            ValueError: Cannot find index of bin containing x = 2.0, which is outside of histogram x-boundaries [0.0, 2.0)
            >>> h.bin_index(-1.0)
            ValueError: Cannot find index of bin containing x = -1.0, which is outside of histogram x-boundaries [0.0, 2.0)

        """
        return _get_bin_index(self.binedges, x, label='x')



    def curve(self, variation=''):
        '''
        Curve representation of histogram
        @variation: if given, return the curve for the variation of this name. Otherwise, return the nominal curve
        '''
        x = np.repeat(self.binedges, 2)[1:-1]
        y = np.repeat(self.heights, 2)
        return (x, y)



    def points(self, variation='', shift=0., abs_shift=False):
        '''
        Point representation of histogram
        If @shift is given, the x-coordinates of the midpoints are given shifted by this absolute x-value (if @abs_shift=True) or relative fraction of the corresponding bin's width (if @abs_shift=False)
        '''
        midx = self.binedges[:-1] + 0.5 * self.binsizes
        if abs_shift:
            midx = midx + shift
        else:
            midx = midx + shift * self.binsizes
        return (midx, self.heights)



    def errorband(self, *args, **kwargs):
        '''
        Basically same as errorbars method, only in curve representation
        @*args and @**kwargs get passed on to self.net_variations()
        '''
        upper, lower = self.net_variations(*args, **kwargs)
        x = np.repeat(self.binedges, 2)[1:-1]

        with warnings.catch_warnings():
            warnings.simplefilter('ignore', RuntimeWarning) # do not warn about NaNs in the division in the following line
            yupper = np.repeat(upper/self.binsizes, 2)
            ylower = np.repeat(lower/self.binsizes, 2)
            return x, yupper, ylower



    def _recalculate_variation_areas_quadsum(self, var_areas, newedges):
        sign = np.sign(np.sum(var_areas) - np.sum(self.areas)) # infer the direction of the shifts
        shifts = var_areas - self.areas
        return np.sqrt(np.add.reduceat(shifts**2, np.searchsorted(self.binedges, newedges)[:-1])) * sign + \
               np.add.reduceat(self.areas, np.searchsorted(self.binedges, newedges)[:-1])



    def rebin(self, newedges):
        '''
        Rebin to @newedges
        Each element of @newedges should correspond to an existing binedge, i.e. only existing bins are merged

        CAUTION: currently ASSUMES that each uncorrelated variation only has shifts in one direction of the nominal
        (i.e. it is either higher or lower everywhere)!
        '''
        if isinstance(newedges, list):
            newedges = np.array(newedges)
        # First altogether drop bins that are outside of the outermost new edges:
        min_index = self.bin_index(np.min(newedges))
        max_index = None if np.max(newedges) == np.max(self.binedges) else self.bin_index(np.max(newedges))
        if min_index > 0 or max_index is not None:
            self.uncorr_variations = {name : areas[min_index:max_index] for name, areas in self.uncorr_variations.items()}
            self.corr_variations = {name : areas[min_index:max_index] for name, areas in self.corr_variations.items()}
            self.areas = self.areas[min_index:max_index]
            self.binedges = self.binedges[min_index:None] if max_index is None else self.binedges[min_index:max_index+1]
        # CAUTION: the order of rebinning variations, binedges, and areas matters here!
        self.uncorr_variations = {name : self._recalculate_variation_areas_quadsum(areas, newedges) for name, areas in self.uncorr_variations.items()}
        self.corr_variations = {name : np.add.reduceat(areas, np.searchsorted(self.binedges, newedges)[:-1]) for name, areas in self.corr_variations.items()}
        self.areas = np.add.reduceat(self.areas, np.searchsorted(self.binedges, newedges)[:-1])
        self.binedges = np.take(self.binedges, np.searchsorted(self.binedges, newedges))



    def merge_bins(self, xmin, xmax):
        '''
        Merge the bins falling into the given x-range into one bin
        '''
        if xmin >= xmax:
            raise RuntimeError('Must have xmin < xmax for merging histogram bins')
        newedges = np.concatenate((self.binedges[self.binedges <= xmin], self.binedges[self.binedges >= xmax]))
        if xmax > np.max(self.binedges): np.append(newedges, np.max(self.binedges)) # so that the highest bin edge stays intact even when xmax is greater than it
        self.rebin(newedges)



    def squash_highest_bin(self, squash_above, new_xmax):
        '''
        Merge all bins from @squash_above upwards and set the highest bin edge to @new_xmax.

        Does nothing if squash_above is >= the highest bin edge.
        '''
        if squash_above >= new_xmax:
            raise RuntimeError('Setting new highest bin edge above the bin squashing threshold is forbidden')
        if squash_above >= np.max(self.binedges):
            return
        self.merge_bins(squash_above, np.max(self.binedges))
        # Only now update binedges:
        self.binedges[-1] = new_xmax



    def height(self, bin_index):
        """Returns the height of the given bin index with uncertainties.

        :returns: height of the indexed bin including its variations
        :rtype: :py:class:`heppy.value`

        Usage example:

        .. code-block:: python

            >>> import heppy as hp
            >>> h = hp.histogram1d([0., 1., 3.], [10., 11.], corr_variations={'systematic__up' : [13., 11.5]})
            >>> v = h.height(1)
            >>> v.nominal
            11.0
            >>> v.corr_variations['systematic__up']
            11.5

        """
        uncorr_variations = {key : (values/self.binwidths)[bin_index] for key, values in self.uncorr_variations.items()}
        corr_variations = {key : (values/self.binwidths)[bin_index] for key, values in self.corr_variations.items()}
        return Value(self.heights[bin_index], uncorr_variations=uncorr_variations, corr_variations=corr_variations)



    def iterheights(self):
        """Generates iterator over heights.

        :returns: bin heights including their variations
        :rtype: :py:class:`heppy.value`

        Usage example:

        .. code-block:: python

            >>> import heppy as hp
            >>> h = hp.histogram1d([0., 1., 3.], [10., 11.], corr_variations={'systematic__up' : [13., 11.5]})
            >>> for height in h.iterheights(): print(height.nominal, height.corr_variations['systematic__up'])
            10.0 13.0
            11.0 11.5

        """
        for bin_index in range(self.nbins):
            yield self.height(bin_index)



    def iterbins(self):
        """Generates iterator over bins, yielding bin edges and heights.

        :returns: bin egdes and nominal bin height
        :rtype: ``tuple`` of the following: ``tuple`` of two ``float``, and one ``float``

        Usage example:

        .. code-block:: python

            >>> import heppy as hp
            >>> h = hp.histogram1d([0., 1., 3.], [10., 11.], corr_variations={'systematic__up' : [13., 11.5]})
            >>> for binedges, height in h.iterbins(): print(binedges, height.nominal)
            (0.0, 1.0) 10.0
            (1.0, 3.0) 11.0
            >>> for binedges, height in h.iterbins(): print(binedges, height.nominal, height.corr_variations['systematic__up'])
            (0.0, 1.0) 10.0 13.0
            (1.0, 3.0) 11.0 11.5

        """
        for bin_lower_edge, bin_upper_edge, height in zip(self.binedges[:-1], self.binedges[1:], self.iterheights()):
            yield (bin_lower_edge, bin_upper_edge), height



    def cumulative(self, side='left', nan_to_num=True):
        """Returns a new histogram containing the cumulative sums of the bin areas.

        Analogous to Numpy's ``cumsum`` function.

        Note: the plot attributes are copied over from ``self`` to the returned
        histogram.


        :param side: which side to integrate from, may be either ``'left'`` or ``'right'``
        :type side: ``str``
        :param nan_to_num: treat NaN as 0 when taking the cumulative sum?
        :type nan_to_num: ``bool``

        :returns: histogram of the cumulative sum of the bin areas
        :rtype: :py:class:`heppy.histogram1d`
        """
        areas = deepcopy(self.areas)
        if nan_to_num:
            areas = np.nan_to_num(areas)
        if side == 'left':
            cumheights = np.cumsum(areas)
        elif side == 'right':
            cumheights = np.cumsum(areas[::-1])[::-1]
        else:
            raise RuntimeError('Argument "side" must be either "left" or "right".')
        return histogram1d(self.binedges, cumheights, areas=False, plot_attributes=self.plot_attributes)



    def to_yoda(self, identifier, metadata={}):
        """Returns the histogram in YODA output format as a string.

        See the websites of `YODA <https://yoda.hepforge.org>`_ and its main user `Rivet <https://rivet.hepforge.org>`_ for more information.

        :param identifier: in-file identifier for the histogram, e.g. ``'/REF/ATLAS_2017_I1614149/d16-x01-y02'``
        :type identifier: ``str``
        :param metadata: optional dictionary of metadata. E.g. for Rivet use, one could have ``metadata = {'IsRef' : 1, 'Path' : '/REF/ATLAS_2017_I1614149/d16-x01-y02', 'Title' : 'doi:10.17182/hepdata.80041.v2/t16'}``
        :type metadata: ``dict``

        :returns: histogram formatted as YODA input string
        :rtype: ``str``

        .. todo::

            Add Variations and ErrorBreakdown fields

            Unit test this method

        """
        template = dedent('''\
        BEGIN YODA_SCATTER2D_V2 {identifier}
        {metadata}
        Type: Scatter2D
        ---
        {data}
        END YODA_SCATTER2D_V2
        ''')
        xval, yval = self.points()
        xerr_minus = xval - self.binedges[:-1]
        xerr_plus = self.binedges[1:] - xval
        yerr_plus, yerr_minus = self.errorbars()
        yerr_plus = np.abs(yerr_plus)
        yerr_minus = np.abs(yerr_minus)
        data_dict = {
            '# xval' : xval,
            'xerr-' : xerr_minus,
            'xerr+' : xerr_plus,
            'yval' : yval,
            'yerr-' : yerr_minus,
            'yerr+' : yerr_plus,
        }
        data = pd.DataFrame.from_dict(data_dict)
        return template.format(
            identifier=identifier,
            metadata='\n'.join(['{key}: {val}'.format(key=key, val=val) for key, val in metadata.items()]),
            data=data.to_string(index=False))



    to_rivet = to_yoda



    def to_root(self, nominal_label='nominal', key_form='{name}_{variation}', errors=None):
        """Returns a dictionary of variation names mapped to ROOT TH1D's.

        :param nominal_label: dictionary key to give to the nominal histogram
        :type nominal_label: ``str``
        :param key_form: template for keys in the returned dictionary. The
            names of the histograms will also be set to these values. It accepts
            the formatting fields `name` (for `self.name`) and variation (for
            the name of the variation)
        :type key_form: ``str``
        :param errors: optional array of bin errors. Must have the same dimensionality
        as the histogram areas. Note that the same errors are used for all variation
        histograms.
        :type errors: ``np.array`` or ``None``

        :returns: dictionary of ROOT histograms (values) mapped by variation
            name (keys)
        :rtype: ``dict`` of ``str`` and ``ROOT.TH1D``
        """
        import ROOT
        import root_numpy as rnp
        from array import array
        def to_th1(name, binedges, areas):
            th1 = ROOT.TH1D(name, name, len(binedges)-1, array('d', [b for b in binedges]))
            rnp.array2hist(areas, th1, errors=errors)
            return th1
        th1s = {}
        # Add nominal
        key = key_form.format(name=self.name, variation=nominal_label)
        th1s[key] = to_th1(key, self.binedges, self.areas)
        for variation, areas in {**self.uncorr_variations, **self.corr_variations}.items():
            key = key_form.format(name=self.name, variation=variation)
            th1s[key] = to_th1(key, self.binedges, areas)
        return th1s



    def to_root_file(self, filename, recreate=False, **kwargs):
        """Writes this histogram to a ROOT file as a set of TH1D's.

        :param filename: name or path to the ROOT file. May exist or not.
        :type filename: ``str``
        :param recreate: only has an effect if the ROOT file exists already. If
            `True`, it will be overwritten. Otherwise, the new histograms will
            be added to the existing file.
        :type recreate: ``bool``
        :param **kwargs: keyword arguments that get passed on to
            :py:func:`histogram1d.to_root`
        """
        import ROOT
        mode = 'RECREATE' if recreate else 'UPDATE'
        rootfile = ROOT.TFile(filename, mode)
        if not rootfile.IsOpen():
            raise OSError(f'Cannot open ROOT file "{filepath}"')
        th1s = self.to_root(**kwargs)
        for key, th1 in th1s.items():
            th1.Write(key)





def zeros_like(a, name=None):
    '''
    Returns a histogram of zeros with the same bin edges and variations as ``a``.

    The attributes and plot attributes are not kept.

    :param a: the bin edges and variation names of ``a`` define these same
    attributes of the returned histogram.
    :type a: :py:class:`heppy.histogram1d` or :py:class:`heppy.histogram2d`
    :param name: if given, this becomes the name attribute of the returned
    histogram. If ``Non``, the returned histogram has the name ``"zeros like " + a.name``
    :type name: ``str`` or ``None``

    :returns: :py:class:`heppy.histogram1d` or :py:class:`heppy.histogram2d`
    '''
    b = deepcopy(a)
    zeros = np.zeros_like(a.areas)
    b.areas = np.zeros_like(a.areas)
    for variation in b.uncorr_variations.keys():
        b.uncorr_variations[variation] = np.zeros_like(a.areas)
    for variation in b.corr_variations.keys():
        b.corr_variations[variation] = np.zeros_like(a.areas)
    b.name = f'zeros like {a.name}' if name is None else name
    b.attributes = {}
    b.plot_attributes = {}
    return b





def from_file(infilename, key):
    '''
    Read histogram written out by heppy (using :py:class:`heppy.basehistogram.to_file`).

    :param infilename: name of the file that the histogram should be read from
    :type infilename: :code:`str`
    :param key: name/key of the histogram inside the input file
    :type key: :code:`str`

    :returns: :py:class:`heppy.histogram1d` or :py:class:`heppy.histogram2d`
    '''
    with open(infilename, 'r') as infile:
        sep = ' : '
        for line in infile:
            if line.split(sep)[0] == key:
                from numpy import array, nan, inf
                return eval(line.split(sep)[1])
        raise RuntimeError('Could not find histogram key "{key}" in input file "{infilename}"'.format(key=key, infilename=infilename))





def histdiv(a, b, corr=None, ignore_denominator_uncertainty=False):
    '''
    Sophisticated division of two histograms

    :param a: numerator histogram
    :type a: :py:class:`heppy.basehistogram`
    :param b: denominator histogram
    :type b: :py:class:`heppy.basehistogram`
    :param corr: information on how a and b are correlated --- NOT YET IMPLEMENTED, do not use
    :param ignore_denominator_uncertainty: switch to ignore the variations of the denominator histogram. If True, divide all variations of the numerator histogram by the nominal denominator histogram.
    :type ignore_denominator_uncertainty: :code:`bool`

    NOTE: the returned ratio histogram's bin heights are not given "per bin size", but take the role that the areas have for histograms that do not represent a ratio.

    :returns: ratio histogram a/b with variations treated as specified
    :raises NotImplementedError: if :code:`corr` is not :code:`None` (remains to be implemented)
    '''
    if ignore_denominator_uncertainty:
        b_no_uncertainty = deepcopy(b)
        b_no_uncertainty.uncorr_variations = {}
        b_no_uncertainty.corr_variations = {}
        r = a / b_no_uncertainty
    else:
        r = a / b
    # Undo division by bin sizes (areas=False):
    r = type(a)(r.binedges, r.areas, areas=False, name=r.name, uncorr_variations=r.uncorr_variations, corr_variations=r.corr_variations, attributes=r.attributes, plot_attributes=r.plot_attributes)
    if isinstance(corr, str) and corr.startswith('subset-toys'):
        ntoys = int(corr.split('-')[-1])
        if (ntoys < 100):
            print('Warning: it is not recommended to use ntoys < 100! Uncertainty propagation will be inaccurate, including due to internal mechanism for dealing with bins in which numerator == denominator!')
        both = a.areas
        denonly = b.areas - a.areas
        if (denonly < 0).any():
            raise RuntimeError('Numerator is not a subset of denominator! Cannot use this uncertainty propagation method.')
        both_toys = np.random.poisson(both, (ntoys, *both.shape))
        denonly_toys = np.random.poisson(denonly, (ntoys, *both.shape))
        ratio_toys = both_toys / (both_toys + denonly_toys)
        #
        # Compute final uncertainty as RMS distance from nominal (up and down separately)
        #
        shift_toys = ratio_toys - r.heights
        upshift_toys = np.clip(shift_toys, 0.0, None)
        downshift_toys = np.clip(shift_toys, None, 0.0)
        # Need the effective number of toys separately for up and down
        # To get it, we just count how many down and how many up fluctuations we had in each bin
        n_effective_toys_up = np.sum(np.ceil(upshift_toys), axis=0)
        n_effective_toys_down = np.sum(np.ceil(np.abs(downshift_toys)), axis=0)
        # Note that in each bin, n_effective_toys_up and n_effective_toys_down sum to either ntoys or zero
        # The latter happens if the numerator and denominator hists have exactly the same entries, and hence
        # will fluctuate coherently and not incur any uncertainty
        # To avoid dividing by zero in such cases, we set them to 1. It doesn't matter what we set them to,
        # since the uncertainties will anyway be zero in these bins.
        n_effective_toys_up = np.clip(n_effective_toys_up, 1.0, None)
        n_effective_toys_down = np.clip(n_effective_toys_down, 1.0, None)
        r.uncorr_variations['Statistical__1up'] = r.areas + np.sqrt(np.sum(upshift_toys**2, axis=0) / n_effective_toys_up) * r.binsizes
        r.uncorr_variations['Statistical__1down'] = r.areas - np.sqrt(np.sum(downshift_toys**2, axis=0) / n_effective_toys_down) * r.binsizes
    elif corr is not None:
        raise NotImplementedError('Division of histograms that are not statistically independent still needs to be implemented.')
    return r





# Convenience alias
histogram = histogram1d





class histogram2d(basehistogram):
    '''
    Heppy two-dimensional histogram.
    This currently has much more limited functionality than the 1D histogram class, although probably
    most (if not all) of the former's mathematical operations should also work for the 2D histogram
    (at least with minor modifications).

    Note: only independent binnings of the two axes are supported (i.e. y-bins don't depend on x-bins and
    vice versa).
    '''
    def __init__(self, *args, **kwargs):
        super(histogram2d, self).__init__(*args, **kwargs)



    @property
    def nbins(self):
        '''
        :returns: tuple of number of bins along x- and y-axis
        '''
        return len(self.binedges[0]) - 1, len(self.binedges[1]) - 1



    def bin_index_x(self, x):
        """Returns the index of the x-axis bin that contains the given x-value.

        Lower bin edges are included in a bin, upper bin edges are excluded (same as in the `ROOT <https://root.cern>`_ convention).

        :param x: x-value
        :type x: ``float``

        :returns: index of x-axis bin that contains the x-value
        :rtype: ``int``

        :raises ValueError: if x-value lies outside of the outer bin edges of the histogram
        """
        return _get_bin_index(self.binedges[0], x, label='x')



    def bin_index_y(self, y):
        """Returns the index of the y-axis bin that contains the given y-value.

        Lower bin edges are included in a bin, upper bin edges are excluded (same as in the `ROOT <https://root.cern>`_ convention).

        :param y: y-value
        :type y: ``float``

        :returns: index of y-axis bin that contains the y-value
        :rtype: ``int``

        :raises ValueError: if y-value lies outside of the outer bin edges of the histogram
        """
        return _get_bin_index(self.binedges[1], y, label='y')



    def points(self):
        '''
        Point representation of 2D histogram.

        This involves flattening/ravelling the histogram bin midpoints and heights to one-dimensional arrays.
        The flattening is done in row-major, C-style order, with the y-axis index changing fastest and the
        x-axis index changing slowest.

        :returns: :code:`tuple` of x-axis bin midpoints, y-axis bin midpoints, and heights
        '''
        binedges_x, binedges_y = self.binedges
        # Calculate bin midpoints in the x and the y direction
        mid_x = binedges_x[:-1] + 0.5 * _calculate_bin_sizes(binedges_x)
        mid_y = binedges_y[:-1] + 0.5 * _calculate_bin_sizes(binedges_y)
        # Now we need to repeat the mid_x and mid_y elements such that they match the
        # raveled heights
        # The y bin changes faster than the x bin
        nbins_x, nbins_y = self.nbins
        expanded_mid_x = np.repeat(mid_x, nbins_y)
        expanded_mid_y = np.tile(mid_y, nbins_x)
        expanded_heights = np.ravel(self.heights)
        return expanded_mid_x, expanded_mid_y, expanded_heights



    def _recalculate_variation_areas_quadsum(self, var_areas, newedges):
        sign = np.sign(np.sum(var_areas) - np.sum(self.areas)) # infer the direction of the shifts
        shifts = var_areas - self.areas
        nominal = self.areas.copy() # Nominal areas to be rebinned along with the variation shifts
        for index in [0, 1]:
            # print(self.binedges[index])
            # print(newedges[index])
            shifts = sign * np.sqrt(np.add.reduceat(shifts**2, np.searchsorted(self.binedges[index], newedges[index])[:-1], axis=index))
            nominal = np.add.reduceat(nominal, np.searchsorted(self.binedges[index], newedges[index])[:-1], axis=index)
        return nominal + shifts



    def rebin(self, newedges):
        '''
        Rebin 2D histogram. Correlated and uncorrelated variations will be recalculated to match the new bin edges.

        CAUTION: currently ASSUMES that each uncorrelated variation only has shifts in one direction of the nominal
        (i.e. it is either higher or lower everywhere)!

        :param newedges: new bin edges. Each new bin edge should correspond to an existing bin edge, i.e. only existing bins are merged
        :type newedges: :code:`tuple` of two :code:`numpy.array`

        :raises: :code:`ValueError` if newedges is not of the correct type
        '''
        if not isinstance(newedges, tuple) or not len(newedges) == 2 or not all([isinstance(e, np.ndarray) for e in newedges]):
            raise ValueError('Argument "newedges" must be a 2-tuple of Numpy arrays')
        # Rebin along one axis after the other --- the order does not matter!
        # (Ensure that the order indeed does not matter when dealing with uncorrelated uncertainties,
        # there may be caveats?)
        self.binedges = list(self.binedges) # make binedges (tuple!) mutable by converting it to a list temporarily...
        # First altogether drop bins that are outside of the outermost new edges:
        for index in [0, 1]:
            bin_index_function = [self.bin_index_x, self.bin_index_y][index]
            min_index = bin_index_function(np.min(newedges[index]))
            max_index = None if np.isclose(np.max(newedges[index]), np.max(self.binedges[index])) else bin_index_function(np.max(newedges[index]))
            if min_index > 0 or max_index is not None:
                if index == 0:
                    self.uncorr_variations = {name : areas[min_index:max_index,:] for name, areas in self.uncorr_variations.items()}
                    self.corr_variations = {name : areas[min_index:max_index,:] for name, areas in self.corr_variations.items()}
                    self.areas = self.areas[min_index:max_index,:]
                else:
                    self.uncorr_variations = {name : areas[:,min_index:max_index] for name, areas in self.uncorr_variations.items()}
                    self.corr_variations = {name : areas[:,min_index:max_index] for name, areas in self.corr_variations.items()}
                    self.areas = self.areas[:,min_index:max_index]
                self.binedges[index] = self.binedges[index][min_index:None] if max_index is None else self.binedges[index][min_index:max_index+1]
        self.uncorr_variations = {name : self._recalculate_variation_areas_quadsum(areas, newedges) for name, areas in self.uncorr_variations.items()}
        for index in [0, 1]:
            # CAUTION: the order of rebinning variations, binedges, and areas matters here!
            self.corr_variations = {name : np.add.reduceat(areas, np.searchsorted(self.binedges[index], newedges[index])[:-1], axis=index) for name, areas in self.corr_variations.items()}
            self.areas = np.add.reduceat(self.areas, np.searchsorted(self.binedges[index], newedges[index])[:-1], axis=index)
            self.binedges[index] = np.take(self.binedges[index], np.searchsorted(self.binedges[index], newedges[index]))
        self.binedges = tuple(self.binedges) # convert binedges back to tuple



    def as_1d(self, name=''):
        '''
        Return a copied one-dimensional reinterpretation of this histogram.
        This only works if the histogram only has one bin in one of its dimensions. This dimension will
        then be ignored.

        :param name: name for the reinterpreted histogram
        :type name: :code:`str`
        '''
        kept_axis = None
        if len(self.binedges[0]) == 2:
            kept_axis = 1
        elif len(self.binedges[1]) == 2:
            kept_axis = 0
        if kept_axis is None:
            raise ValueError('Cannot reinterpret 2D histogram as 1D histogram, since neither of its axes has only one bin')
        binedges = self.binedges[kept_axis]
        areas = self.areas.flatten(order='C') # or 'F'
        cv = {key : areas.flatten(order='C') for key, areas in self.corr_variations.items()}
        uv = {key : areas.flatten(order='C') for key, areas in self.uncorr_variations.items()}
        return histogram1d(binedges, areas, areas=True, name=name, corr_variations=cv, uncorr_variations=uv, attributes=self.attributes, plot_attributes=self.plot_attributes)



    def project(self, axis, name=''):
        '''
        Project histogram to one axis by integrating over the other. Correlated and uncorrelated
        uncertainties are computed for the resulting one-dimensional histogram.

        :param axis: which axis to project onto, i.e. the *axis that is kept*
        :type axis: :code:`'x'` or :code:`'y'`

        :param name: name for the projection histogram
        :type name: :code:`str`


        :returns: :py:class:`heppy.histogram1d` representing the projection

        :raises: :code:`ValueError` if invalid axis identifier is given
        '''
        if not axis in ['x', 'y']:
            raise ValueError('Invalid axis "{axis}", valid choices are: "x", "y"'.format(axis=axis))
        binedge_index = {'x' : 0, 'y' : 1}[axis]
        # Rebin the histogram to (2D) new edges such that there is only one bin in the projected-out
        # dimension:
        newedges_x = self.binedges[0] if axis == 'x' else np.array([self.binedges[0][0], self.binedges[0][-1]])
        newedges_y = self.binedges[1] if axis == 'y' else np.array([self.binedges[1][0], self.binedges[1][-1]])
        newedges = (newedges_x, newedges_y)
        projected = deepcopy(self)
        projected.rebin(newedges)
        if name == '':
            name = self.name + ' projected to {axis}-axis'.format(axis=axis)
        return projected.as_1d(name=name)



    def slice(self, axis, bin_index, name=''):
        """Returns 1D histogram of the distribution along one axis in a given bin of the other axis.

        :param axis: axis along which the slicing is done, i.e. the *axis that is kept*
        :type axis: :code:`'x'` or :code:`'y'`

        :param bin_index: index of the bin on the *axis that is not kept*
        :type bin_index: ``int``

        :param name: name for the slice histogram
        :type name: ``str``


        :returns: 1D histogram of the slice
        :rtype: :py:class:`heppy.histogram1d`
        """
        if not axis in ['x', 'y']:
            raise ValueError('Invalid axis "{axis}", valid choices are: "x", "y"'.format(axis=axis))
        other_axis = {'x' : 'y', 'y' : 'x'}[axis]
        binedge_index = {'x' : 0, 'y' : 1}[other_axis]
        bin_lower_edge = self.binedges[binedge_index][bin_index]
        bin_upper_edge = self.binedges[binedge_index][bin_index+1]
        # The actual slicing happens below.
        # The strategy is to rebin a copy of the 2D histogram such that the bins in the
        # ``axis``-direction are kept while only one bin in the ``other_axis``-direction
        # is kept. The resulting 2D histogram is then cast to a 1D histogram.
        sliced = deepcopy(self)
        newedges = {
            'x' : (sliced.binedges[0], np.array([bin_lower_edge, bin_upper_edge])),
            'y' : (np.array([bin_lower_edge, bin_upper_edge]), sliced.binedges[1]),
            }[axis]
        sliced.rebin(newedges)
        width_of_slice_bin = bin_upper_edge - bin_lower_edge
        sliced = sliced * (1.0 / width_of_slice_bin) # is this correct? It makes the test fail --- maybe the test is wrong? How do we want to normalise?
        if name == '':
            name = self.name + ' slice along {axis}-axis in {other_axis}-bin [{binmin}, {binmax})'.format(axis=axis, other_axis=other_axis, binmin=bin_lower_edge, binmax=bin_upper_edge)
        return sliced.as_1d(name=name)



    def height(self, bin_index_x, bin_index_y):
        """Returns the height of the given bin indices with uncertainties.

        :param bin_index_x: bin index along x-axis
        :param bin_index_x: ``int``
        :param bin_index_y: bin index along y-axis
        :param bin_index_y: ``int``

        :returns: height of the indexed bin including its variations
        :rtype: :py:class:`heppy.value`
        """
        uncorr_variations = {key : (values/self.binsizes)[bin_index_x,bin_index_y] for key, values in self.uncorr_variations.items()}
        corr_variations = {key : (values/self.binsizes)[bin_index_x,bin_index_y] for key, values in self.corr_variations.items()}
        return Value(self.heights[bin_index_x,bin_index_y], uncorr_variations=uncorr_variations, corr_variations=corr_variations)



    def iterheights(self, faster='y'):
        """Generates iterator over heights.

        :param faster: controls the iteration order by specifying along which axis the bin index changes faster
        :type faster: ``str``; ``'x'`` or ``'y'``

        :returns: bin heights including their variations
        :rtype: :py:class:`heppy.value`
        """
        if not faster in ('x', 'y'):
            raise ValueError('Valid values for keyword argument "faster" are: "x", "y"')
        nbins_x, nbins_y = self.nbins
        if faster == 'y':
            for bin_index_x in range(nbins_x):
                for bin_index_y in range(nbins_y):
                    yield self.height(bin_index_x, bin_index_y)
        elif faster == 'x':
            for bin_index_y in range(nbins_y):
                for bin_index_x in range(nbins_x):
                    yield self.height(bin_index_x, bin_index_y)



    def iterbins(self):
        """Generates iterator over bins, yielding bin edges and heights.

        :returns: x-axis bin egdes, y-axis bin edges, and bin height with variations
        :rtype: ``tuple`` of the following: ``tuple`` of two ``float``, ``tuple`` of two ``float``, and one :py:class:`heppy.value`

        Usage example:

        .. code-block:: python

            >>> import heppy as hp
            >>> import numpy as np
            >>> heights = np.array([             # bin heights
                    [1., 5.],
                    [2., 6.],
                    [3., 7.],
                    ])
            >>> x = np.array([-7., 0., 5., 50.]) # bin edges in x
            >>> y = np.array([-1., 0., 1.])      # bin edges in y
            >>> h = hp.histogram2d((x, y), heights)
            >>> for binedges_x, binedges_y, height in h.iterbins(): print(binedges_x, binedges_y, height.nominal)
            (-7.0, 0.0) (-1.0, 0.0) 1.0
            (-7.0, 0.0) (0.0, 1.0) 5.0
            (0.0, 5.0) (-1.0, 0.0) 2.0
            (0.0, 5.0) (0.0, 1.0) 6.0
            (5.0, 50.0) (-1.0, 0.0) 3.0
            (5.0, 50.0) (0.0, 1.0) 7.0

        """
        nbins_x, nbins_y = self.nbins
        binedges_x, binedges_y = self.binedges
        bin_lower_edges_x = np.repeat(binedges_x[:-1], nbins_y)
        bin_upper_edges_x = np.repeat(binedges_x[1:], nbins_y)
        bin_lower_edges_y = np.tile(binedges_y[:-1], nbins_x)
        bin_upper_edges_y = np.tile(binedges_y[1:], nbins_x)
        for bin_lower_edge_x, bin_upper_edge_x, bin_lower_edge_y, bin_upper_edge_y, height in zip(bin_lower_edges_x, bin_upper_edges_x, bin_lower_edges_y, bin_upper_edges_y, self.iterheights(faster='y')):
            yield (bin_lower_edge_x, bin_upper_edge_x), (bin_lower_edge_y, bin_upper_edge_y), height



    def to_root(self, nominal_label='nominal', key_form='{name}_{variation}', errors=None):
        """Returns a dictionary of variation names mapped to ROOT TH2D's.

        :param nominal_label: dictionary key to give to the nominal histogram
        :type nominal_label: ``str``
        :param key_form: template for keys in the returned dictionary. The
            names of the histograms will also be set to these values. It accepts
            the formatting fields `name` (for `self.name`) and variation (for
            the name of the variation)
        :type key_form: ``str``
        :param errors: optional array of bin errors. Must have the same dimensionality
        as the histogram areas. Note that the same errors are used for all variation
        histograms.
        :type errors: ``np.array`` or ``None``


        :returns: dictionary of ROOT histograms (values) mapped by variation
            name (keys)
        :rtype: ``dict`` of ``str`` and ``ROOT.TH2D``
        """
        import ROOT
        import root_numpy as rnp
        from array import array
        def to_th2(name, binedges, areas):
            th2 = ROOT.TH2D(name, name, len(binedges[0])-1, array('d', [b for b in binedges[0]]), len(binedges[1])-1, array('d', [b for b in binedges[1]]))
            rnp.array2hist(areas, th2, errors=errors)
            return th2
        th2s = {}
        # Add nominal
        key = key_form.format(name=self.name, variation=nominal_label)
        th2s[key] = to_th2(key, self.binedges, self.areas)
        for variation, areas in {**self.uncorr_variations, **self.corr_variations}.items():
            key = key_form.format(name=self.name, variation=variation)
            th2s[key] = to_th2(key, self.binedges, areas)
        return th2s



    def to_root_file(self, filename, recreate=False, **kwargs):
        """Writes this histogram to a ROOT file as a set of TH2D's.

        :param filename: name or path to the ROOT file. May exist or not.
        :type filename: ``str``
        :param recreate: only has an effect if the ROOT file exists already. If
            `True`, it will be overwritten. Otherwise, the new histograms will
            be added to the existing file.
        :type recreate: ``bool``
        :param **kwargs: keyword arguments that get passed on to
            :py:func:`histogram2d.to_root`
        """
        import ROOT
        mode = 'RECREATE' if recreate else 'UPDATE'
        rootfile = ROOT.TFile(filename, mode)
        if not rootfile.IsOpen():
            raise OSError(f'Cannot open ROOT file "{filepath}"')
        th2s = self.to_root(**kwargs)
        for key, th2 in th2s.items():
            th2.Write(key)



class histostack(object):
    '''
    Stack of one-dimensional histograms

    :param histograms: histograms in the stack
    :type histograms: :code:`list` of :py:class:`heppy.histogram1d`
    :param attributes: dictionary of completely arbitrary attributes that the user can provide/change/access. E.g. information on how to plot
    :type attributes: :code:`dict`
    '''
    def __init__(self, histograms, attributes={}):
        super(histostack, self).__init__()
        self.histograms = histograms
        self.attributes = attributes

    @property
    def total(self):
        '''
        :returns: :py:class:`heppy.histogram1d` that is the combination of all the stacked ones, with the combined uncertainty. If the stack has no histograms, returns :code:`None`
        '''
        if not self.histograms: return None
        try:
            return sum(self.histograms[1:], self.histograms[0])
        except IndexError:
            return self.histograms[0]

    def iterbands(self):
        '''
        Bands are 3-tuples of (curve representation) x-values as well as two subsequent curves that are useful as arguments to :code:`plt.fill_between()`.
        E.g.:

        .. code-block:: python

            for histogram, band in zip(stack.histograms, stack.iterbands()):
                ax.fill_between(*band, **histogram.attributes)

        The bands are ordered such that the first histogram in the stack is at the top and the last at the bottom
        '''
        if not self.histograms:
            return
        (x, y) = self.histograms[0].curve()
        bottom = np.zeros_like(y)
        bands = []
        for histo in reversed(self.histograms):
            top = bottom + histo.curve()[1]
            bands.append((x, bottom, top))
            bottom = top
        for band in reversed(bands): # Storing the bands first and looping over them again in reversed order means that the topmost histogram has the first colour
            yield band



#!/usr/bin/env python3

import numpy as np
import heppy as hp
import fnmatch



def get_TH(rootfile, histopath):
    '''
    @rootfile may be a ROOT.TFile opened in read mode or the path to an existing ROOT file
    @histopath is the path of the histogram inside the file
    '''
    import ROOT
    ROOT.PyConfig.IgnoreCommandLineOptions = True
    if isinstance(rootfile, str):
        filepath = rootfile
        rootfile = ROOT.TFile(rootfile, 'r') # open ROOT file
        if not rootfile.IsOpen():
            raise OSError('Cannot open ROOT file "{filepath}"'.format(filepath=filepath))
    histogram_owned_by_file = rootfile.Get(histopath)
    ROOT.gROOT.cd()
    try:
        return histogram_owned_by_file.Clone()
    except ReferenceError:
        raise ReferenceError('Cannot read histogram "{histopath}" in file {filepath}'.format(histopath=histopath, filepath=rootfile.GetName()))



def readroot_binareas(rootfile, histopath, rebin=False):
    '''
    Read bin areas only and return them as a numpy array
    '''
    import root_numpy as rnp
    th = get_TH(rootfile, histopath)
    if rebin:
        th = th.Rebin(rebin)
    return rnp.hist2array(th)



def readroot_binedges1d(th1):
    '''
    Read the binedges of a TH1 histogram and return them as a one-dimensional Numpy array
    '''
    return np.array([th1.GetBinLowEdge(i) for i in range(1, th1.GetNbinsX() + 2)])



def readroot_binedges2d(th2):
    '''
    Read the binedges of a TH2 histogram and return them as a 2-tuple of one-dimensional Numpy arrays,
    where the first element represents the binedges along the x-axis and the second the binedges along
    the y-axis
    '''
    x_axis = th2.GetXaxis()
    x_edges = np.array([x_axis.GetBinLowEdge(i) for i in range(1, x_axis.GetNbins() + 2)])
    y_axis = th2.GetYaxis()
    y_edges = np.array([y_axis.GetBinLowEdge(i) for i in range(1, y_axis.GetNbins() + 2)])
    return x_edges, y_edges



def read_unix_wildcard_corr_variations(filepath, variation_paths, rebin=False):
    '''
    WARNING: slow in general.
    '''
    import uproot
    corr_variations = {}
    urfile = uproot.open(filepath)
    all_paths_in_rootfile = [key.decode().split(';')[0] for key in urfile.allkeys()]
    for name_pattern, path_pattern in variation_paths.items():
        matched_paths = fnmatch.filter(all_paths_in_rootfile, path_pattern)
        # print(name_pattern, path_pattern, '{0} matches found!'.format(len(matched_paths)))
        for matched_path in matched_paths:
            # Now to find the matched name. This will be the ROOT directory or object name that matches the name pattern
            matched_name_hits = fnmatch.filter(matched_path.split('/'), name_pattern)
            if not matched_name_hits:
                raise RuntimeError('Could not infer variation name corresponding to variation histogram "{path}": the variation name pattern "{name_pattern}" does not match any of the directory/object names in that path.'.format(path=matched_path, name_pattern=name_pattern))
            if len(matched_name_hits) > 1:
                raise RuntimeError('Encountered ambiguity when infering variation name corresponding to variation histogram "{path}": the variation name pattern "{name_pattern}" matches multiple directory/object names in that path.'.format(path=matched_path, name_pattern=name_pattern))
            matched_name = matched_name_hits[0]
            # print(matched_name + ':\t' + matched_path)
            corr_variations[matched_name] = readroot_binareas(filepath, matched_path, rebin=rebin)
    return corr_variations



def read_corr_variations(filepath, variation_paths, rebin=False, ignore_missing_variations=False):
    '''
    Return dictionary of variations located in @rootfile at @variation_paths (that is, paths inside the ROOT file),
    possibly rebinned by @rebin

    Custom wildcards:
        '@' gets replaced by 0, 1, ..., until no more histograms found
        '#' gets replaced by 1, 2, ..., until no more histograms found

    The variation_paths may also contain Unix-style wildcards (see Python module fnmatch!), in which case all encountered
    paths matching the pattern(s) are included in the returned dictionary.
    CAUTION: no error is raised if no paths match a given wildcard.

    CAUTION: mixing different types of wildcard ('@', '#', Unix-style) gives [deterministic but] undefined behaviour!
             Please don't do this.
    '''
    # Split paths into ones containing different types of wildcards and direct paths:
    unix_wildcard_paths = {}
    indexed_from_zero_paths = {}
    indexed_from_one_paths = {}
    direct_paths = {}
    for name, path in variation_paths.items():
        if any(wildcard_indicator in path for wildcard_indicator in ['*', '?', '[']):
            unix_wildcard_paths[name] = path
        elif '@' in path:
            indexed_from_zero_paths[name] = path
        elif '#' in path:
            indexed_from_one_paths[name] = path
        else:
            direct_paths[name] = path
    # First read paths not containing wildcards:
    corr_variations = {}
    for name, path in direct_paths.items():
        try:
            corr_variations[name] = readroot_binareas(filepath, path, rebin=rebin)
        except ReferenceError:
            if not ignore_missing_variations:
                raise
    # corr_variations = {name : readroot_binareas(filepath, path, rebin=rebin) for name, path in direct_paths.items()}

    for starting_index, paths_dictionary in zip([0, 1], [indexed_from_zero_paths, indexed_from_one_paths]):
        for name_pattern, path_pattern in paths_dictionary.items():
            i = starting_index
            while True:
                try:
                    name = name_pattern.replace('@', str(i))
                    path = path_pattern.replace('@', str(i))
                    corr_variations[name] = readroot_binareas(filepath, path, rebin=rebin)
                except:
                    break
                i += 1

    if (unix_wildcard_paths):
        # This is only executed if we have any Unix wildcard paths, because using uproot/matching wildcards may be very slow!
        return {**corr_variations, **read_unix_wildcard_corr_variations(filepath, unix_wildcard_paths, rebin=rebin)}
    return corr_variations



def _readroot2d(th, rootfile, histopath, rebin=False, variation_paths={}, ignore_missing_variations=False, **kwargs):
    '''
    @rootfile may be a ROOT.TFile opened in read mode or the path to an existing ROOT file
    @histopath is the path of the histogram inside the file
    @rebin, if given, is passed to ROOT.TH2.Rebin()
    @variation_paths is a dictionary of variation names (keys) and paths to the variation histograms inside @rootfile (values).
                     The @variation paths may contain the following wildcards:
                         - "@" will be interpreted as a numerical index starting at 0 and continuing until no more histograms are found
                         - "#" is the same as "@", but starting at 1
                         - Arbitrary Unix-style wildcards (see Python module fnmatch). These may not be combined with the above
                           custom wildcards. NOTE: including Unix-style wildcards may be very SLOW and introduces a dependency on
                           uproot. If either of these poses a problem, contact the Heppy maintainers, there's a lot of room for
                           speed optimisation here, and the uproot dependency can in principle be circumvented.
                     CAUTION: no error is raised if no paths match a given wildcard (@, @1, ), BUT if a path not containing wildcards
                              is matched, this DOES raise an error.
    @**kwargs get passed on to histogram constructor. An important one is @areas=False (set to True in _readroot2d by default!) if retrieving ratios
              (e.g. efficiencies), and it may be convenient to pass @name=<something> (default: <ROOT histogram>.GetName())
    '''
    import ROOT
    ROOT.PyConfig.IgnoreCommandLineOptions = True
    import root_numpy as rnp
    if rebin:
        th = th.Rebin(rebin)
    binedges = readroot_binedges2d(th)
    areas_nominal = rnp.hist2array(th)
    areas_up = np.copy(areas_nominal)
    areas_down = np.copy(areas_nominal)
    for i in range(areas_nominal.shape[0]):
        for j in range(areas_nominal.shape[1]):
            stat_uncert = th.GetBinError(i + 1, j + 1)
            areas_up[i,j] += stat_uncert
            areas_down[i,j] -= stat_uncert
    filepath = rootfile.GetName() if isinstance(rootfile, ROOT.TFile) else rootfile
    attributes = {
        'name' : th.GetName(),
        'title' : th.GetTitle(),
        'provenance' : '{filepath}:{histopath}'.format(filepath=filepath, histopath=histopath),
    }
    corr_variations = read_corr_variations(filepath, variation_paths, rebin=rebin, ignore_missing_variations=ignore_missing_variations)
    if not 'areas' in kwargs:
        kwargs['areas'] = True
    if not 'name' in kwargs:
        kwargs['name'] = th.GetName()
    return hp.histogram2d(binedges, areas_nominal, uncorr_variations={'Statistical__1up' : areas_up, 'Statistical__1down' : areas_down}, corr_variations=corr_variations, attributes=attributes, plot_attributes={'label' : kwargs['name']}, **kwargs)



def _readroot1d(th, rootfile, histopath, rebin=False, variation_paths={}, ignore_missing_variations=False, **kwargs):
    '''
    @rootfile may be a ROOT.TFile opened in read mode or the path to an existing ROOT file
    @histopath is the path of the histogram inside the file
    @rebin, if given, is passed to ROOT.TH1.Rebin()
    @variation_paths is a dictionary of variation names (keys) and paths to the variation histograms inside @rootfile (values).
                     The @variation paths may contain the following wildcards:
                         - "@" will be interpreted as a numerical index starting at 0 and continuing until no more histograms are found
                         - "#" is the same as "@", but starting at 1
                         - Arbitrary Unix-style wildcards (see Python module fnmatch). These may not be combined with the above
                           custom wildcards. NOTE: including Unix-style wildcards may be very SLOW and introduces a dependency on
                           uproot. If either of these poses a problem, contact the Heppy maintainers, there's a lot of room for
                           speed optimisation here, and the uproot dependency can in principle be circumvented.
                     CAUTION: no error is raised if no paths match a given wildcard (@, @1, ), BUT if a path not containing wildcards
                              is matched, this DOES raise an error.
    @**kwargs get passed on to histogram constructor. An important one is @areas=False (set to True in _readroot1d by default!) if retrieving ratios
              (e.g. efficiencies), and it may be convenient to pass @name=<something> (default: <ROOT histogram>.GetName())
    '''
    import ROOT
    ROOT.PyConfig.IgnoreCommandLineOptions = True
    import root_numpy as rnp
    if rebin:
        th = th.Rebin(rebin)
    binedges = readroot_binedges1d(th)
    areas_nominal = rnp.hist2array(th)
    areas_up = np.copy(areas_nominal)
    areas_down = np.copy(areas_nominal)
    for i in range(len(areas_nominal)):
        stat_uncert = th.GetBinError(i + 1)
        areas_up[i] += stat_uncert
        areas_down[i] -= stat_uncert
    filepath = rootfile.GetName() if isinstance(rootfile, ROOT.TFile) else rootfile
    attributes = {
        'name' : th.GetName(),
        'title' : th.GetTitle(),
        'provenance' : '{filepath}:{histopath}'.format(filepath=filepath, histopath=histopath),
    }
    corr_variations = read_corr_variations(filepath, variation_paths, rebin=rebin, ignore_missing_variations=ignore_missing_variations)
    if not 'areas' in kwargs:
        kwargs['areas'] = True
    if not 'name' in kwargs:
        kwargs['name'] = th.GetName()
    return hp.histogram1d(binedges, areas_nominal, uncorr_variations={'Statistical__1up' : areas_up, 'Statistical__1down' : areas_down}, corr_variations=corr_variations, attributes=attributes, plot_attributes={'label' : kwargs['name']}, **kwargs)



def readroot(rootfile, histopath, rebin=False, variation_paths={}, ignore_missing_variations=False, **kwargs):
    import ROOT
    ROOT.PyConfig.IgnoreCommandLineOptions = True
    th = get_TH(rootfile, histopath)
    # CAUTION: the order of the if statements matters here, because in ROOT, TH2 inherits from TH1!!!
    if isinstance(th, ROOT.TH2):
        return _readroot2d(th, rootfile, histopath, rebin=rebin, variation_paths=variation_paths, ignore_missing_variations=ignore_missing_variations, **kwargs)
    if isinstance(th, ROOT.TH1):
        return _readroot1d(th, rootfile, histopath, rebin=rebin, variation_paths=variation_paths, ignore_missing_variations=ignore_missing_variations, **kwargs)
    raise TypeError('Can only read ROOT.TH1 and ROOT.TH2 objects, but found {type} object in "{rootfile}" at "{histopath}"'.format(type=type(th), rootfile=rootfile, histopath=histopath))



# Auxiliary modules
import pytest
from os.path import join, dirname
import numpy as np
from copy import deepcopy
import sys

# Module to be tested
sys.path.append(join(dirname(__file__), '..'))
import heppy



#
# Set up
#
_areas = np.array([
	[1., 5.],
	[2., 6.],
	[3., 7.],
	])
_binedges_x = np.array([-7., 0., 5., 50.])
_binedges_y = np.array([-1., 0., 1.])
_binedges = (_binedges_x, _binedges_y)
_binsizes = np.array([
	[7.,  7. ],
	[5.,  5. ],
	[45., 45.],
	])
_corr_variations = {
	'corr__1up' : _areas * 1.2,
	'corr__1down' : _areas * 0.8,
}
_uncorr_variations = {
	'uncorr__1up' : _areas * 1.3,
	'uncorr__1down' : _areas * 0.7,
}
_h = heppy.histogram2d(_binedges, _areas, areas=True, corr_variations=_corr_variations, uncorr_variations=_uncorr_variations, name='foo')





def test_calculate_bin_sizes():
	'''
	Test free function used for calculating bin sizes, in the case of a 2D histogram
	'''
	from heppy.histogram import _calculate_bin_sizes
	binsizes = _calculate_bin_sizes(_binedges)
	assert binsizes.shape == (3, 2)
	np.testing.assert_array_almost_equal( binsizes, _binsizes )



def test_binareas():
	'''
	Test computing bin areas of a 2D histogram
	'''
	binsizes = _h.binsizes
	assert binsizes.shape == (3, 2)
	np.testing.assert_array_almost_equal( binsizes, _binsizes )



def test_areas():
	'''
	Test retrieving areas of a 2D histogram
	'''
	areas = _h.areas
	assert areas.shape == (3, 2)
	np.testing.assert_array_almost_equal( areas, _areas )



def test_corr_variations():
	'''
	Test retrieving correlated variations (areas)
	'''
	np.testing.assert_array_almost_equal( _h.corr_variations['corr__1up'], _areas * 1.2 )
	np.testing.assert_array_almost_equal( _h.corr_variations['corr__1down'], _areas * 0.8 )



def test_uncorr_variations():
	'''
	Test retrieving uncorrelated variations (areas)
	'''
	np.testing.assert_array_almost_equal( _h.uncorr_variations['uncorr__1up'], _areas * 1.3 )
	np.testing.assert_array_almost_equal( _h.uncorr_variations['uncorr__1down'], _areas * 0.7 )



def test_heights():
	'''
	Test retrieving heights of a 2D histogram
	'''
	heights = _h.heights
	assert heights.shape == (3, 2)
	np.testing.assert_array_almost_equal( heights, _areas/_binsizes )



def test_nbins():
	'''
	Test retrieving the numbers of bins of a 2D histogram
	'''
	nbins = _h.nbins
	assert nbins == (3, 2)



def test_find_bin():
    # Check cases where x lies within the histogram x-boundaries:
    assert _h.bin_index_x(-7.0) == 0
    assert _h.bin_index_x(-5.0) == 0
    assert _h.bin_index_x(0.) == 1
    assert _h.bin_index_x(0.5) == 1
    assert _h.bin_index_x(5.3) == 2
    # Check cases where x lies outside of the histogram x-boundaries:
    with pytest.raises(ValueError) as exc_info:
        assert _h.bin_index_x(-7.1)
    assert exc_info.value.args[0] == 'Cannot find index of bin containing x = -7.1, which is outside of histogram x-boundaries [-7.0, 50.0)'
    with pytest.raises(ValueError) as exc_info:
        assert _h.bin_index_x(60.)
    assert exc_info.value.args[0] == 'Cannot find index of bin containing x = 60.0, which is outside of histogram x-boundaries [-7.0, 50.0)'
    with pytest.raises(ValueError) as exc_info:
        assert _h.bin_index_x(50.) # IMPORTANT: upper bin edge is NOT part of bin!
    assert exc_info.value.args[0] == 'Cannot find index of bin containing x = 50.0, which is outside of histogram x-boundaries [-7.0, 50.0)'
    # Check cases where y lies within the histogram y-boundaries:
    assert _h.bin_index_y(-1.0) == 0
    assert _h.bin_index_y(-0.5) == 0
    assert _h.bin_index_y(0.) == 1
    assert _h.bin_index_y(0.5) == 1
    # Check cases where y lies outside of the histogram y-boundaries:
    with pytest.raises(ValueError) as exc_info:
        assert _h.bin_index_y(-1.1)
    assert exc_info.value.args[0] == 'Cannot find index of bin containing y = -1.1, which is outside of histogram y-boundaries [-1.0, 1.0)'
    with pytest.raises(ValueError) as exc_info:
        assert _h.bin_index_y(5.)
    assert exc_info.value.args[0] == 'Cannot find index of bin containing y = 5.0, which is outside of histogram y-boundaries [-1.0, 1.0)'



def test_points():
	'''
	Test making the point representation of a 2D histogram
	'''
	midx, midy, heights = _h.points()
	assert midx.shape == midy.shape
	assert midx.shape == heights.shape
	np.testing.assert_array_almost_equal( midx, np.array([-3.5, -3.5, 2.5, 2.5, 27.5, 27.5]) )
	np.testing.assert_array_almost_equal( midy, np.array([-0.5, 0.5, -0.5, 0.5, -0.5, 0.5]) )
	np.testing.assert_array_almost_equal( heights, np.ravel(_areas/_binsizes) )



def test_rebin():
	'''
	Test rebinning of 2D histogram (including corr. and uncorr. variations)
	'''
	newedges_x = np.array([-7., 5., 50.])
	newedges_y = np.array([-1., 1.])
	newedges = (newedges_x, newedges_y)
	h = deepcopy(_h)
	h.rebin(newedges)
	assert isinstance(h.binedges, tuple)
	assert h.binsizes.shape == (2, 1)
	np.testing.assert_array_almost_equal(h.binedges[0], newedges_x)
	np.testing.assert_array_almost_equal(h.binedges[1], newedges_y)
	np.testing.assert_array_almost_equal(h.areas, np.array([[14.], [10.]]))
	np.testing.assert_array_almost_equal(h.corr_variations['corr__1up'], np.array([[14.], [10.]]) * 1.2 )
	np.testing.assert_array_almost_equal(h.corr_variations['corr__1down'], np.array([[14.], [10.]]) * 0.8 )
	np.testing.assert_array_almost_equal(h.uncorr_variations['uncorr__1up'], np.array([[14. + np.sqrt(0.09 + 2.25 + 0.36 + 3.24)], [10. + np.sqrt(0.81 + 4.41)]]) )
	np.testing.assert_array_almost_equal(h.uncorr_variations['uncorr__1down'], np.array([[14. - np.sqrt(0.09 + 2.25 + 0.36 + 3.24)], [10. - np.sqrt(0.81 + 4.41)]]) )



def test_lossy_rebin():
	"""
    Rebinning less than the complete bin range of the histogram.
    Information falling outside of the new outer bin edges is dropped.
    """
	newedges_x = np.array([-7., 5.0])
	newedges_y = np.array([0, 1.])
	newedges = (newedges_x, newedges_y)
	h = deepcopy(_h)
	h.rebin(newedges)
	assert isinstance(h.binedges, tuple)
	assert h.binsizes.shape == (1, 1)
	np.testing.assert_array_almost_equal(h.binedges[0], newedges_x)
	np.testing.assert_array_almost_equal(h.binedges[1], newedges_y)
	np.testing.assert_array_almost_equal(h.areas, np.array([[11.]]))
	np.testing.assert_array_almost_equal(h.corr_variations['corr__1up'], np.array([[11.]]) * 1.2 )
	np.testing.assert_array_almost_equal(h.corr_variations['corr__1down'], np.array([[11.]]) * 0.8 )
	np.testing.assert_array_almost_equal(h.uncorr_variations['uncorr__1up'], np.array([[11.+np.sqrt((5 * 0.3)**2 + (6 * 0.3)**2)]]) )
	np.testing.assert_array_almost_equal(h.uncorr_variations['uncorr__1down'], np.array([[11.-np.sqrt((5 * 0.3)**2 + (6 * 0.3)**2)]]) )



def test_project_x():
	'''
	Test projecting a 2D histogram (including corr. and uncorr. variations) onto the x-axis
	'''
	h = _h.project('x', name='x projection')
	assert isinstance(h, heppy.histogram1d)
	assert h.name == 'x projection'
	np.testing.assert_array_almost_equal(h.binedges, _binedges_x)
	np.testing.assert_array_almost_equal(h.areas, [6., 8., 10.])
	np.testing.assert_array_almost_equal(h.corr_variations['corr__1up'], np.array([6., 8., 10.]) * 1.2 )
	np.testing.assert_array_almost_equal(h.corr_variations['corr__1down'], np.array([6., 8., 10.]) * 0.8 )
	np.testing.assert_array_almost_equal(h.uncorr_variations['uncorr__1up'], np.array([6. + np.sqrt(0.09 + 2.25), 8. + np.sqrt(0.36 + 3.24), 10. + np.sqrt(0.81 + 4.41)]) )
	np.testing.assert_array_almost_equal(h.uncorr_variations['uncorr__1down'], np.array([6. - np.sqrt(0.09 + 2.25), 8. - np.sqrt(0.36 + 3.24), 10. - np.sqrt(0.81 + 4.41)]) )




def test_project_y():
	'''
	Test projecting a 2D histogram (including corr. and uncorr. variations) onto the y-axis
	'''
	h = _h.project('y', name='y projection')
	assert isinstance(h, heppy.histogram1d)
	assert h.name == 'y projection'
	np.testing.assert_array_almost_equal(h.binedges, _binedges_y)
	np.testing.assert_array_almost_equal(h.areas, [6., 18.])
	np.testing.assert_array_almost_equal(h.corr_variations['corr__1up'], np.array([6., 18.]) * 1.2 )
	np.testing.assert_array_almost_equal(h.corr_variations['corr__1down'], np.array([6., 18.]) * 0.8 )
	np.testing.assert_array_almost_equal(h.uncorr_variations['uncorr__1up'], np.array([6. + np.sqrt(0.09 + 0.36 + 0.81), 18. + np.sqrt(2.25 + 3.24 + 4.41)]) )
	np.testing.assert_array_almost_equal(h.uncorr_variations['uncorr__1down'], np.array([6. - np.sqrt(0.09 + 0.36 + 0.81), 18. - np.sqrt(2.25 + 3.24 + 4.41)]) )



def test_slice_x():
    '''
    Test taking a slice of the histogram along the x-axis in a single y-bin
    '''
    h = _h.slice('x', 0)
    assert isinstance(h, heppy.histogram1d)
    assert h.name == 'foo slice along x-axis in y-bin [-1.0, 0.0)'
    np.testing.assert_array_almost_equal(h.binedges, _binedges_x)
    np.testing.assert_array_almost_equal(h.areas, [1., 2., 3])
    np.testing.assert_array_almost_equal(h.corr_variations['corr__1up'], h.areas*1.2 )
    np.testing.assert_array_almost_equal(h.corr_variations['corr__1down'], h.areas*0.8 )
    np.testing.assert_array_almost_equal(h.uncorr_variations['uncorr__1up'], h.areas*1.3 )
    np.testing.assert_array_almost_equal(h.uncorr_variations['uncorr__1down'], h.areas*0.7 )
    h = _h.slice('x', 1, name='bla')
    assert h.name == 'bla'



def test_slice_y():
    '''
    Test taking a slice of the histogram along the x-axis in a single y-bin
    '''
    h = _h.slice('y', 2)
    assert isinstance(h, heppy.histogram1d)
    assert h.name == 'foo slice along y-axis in x-bin [5.0, 50.0)'
    np.testing.assert_array_almost_equal(h.binedges, _binedges_y)
    np.testing.assert_array_almost_equal(h.areas, [3., 7.])
    np.testing.assert_array_almost_equal(h.corr_variations['corr__1up'], h.areas*1.2 )
    np.testing.assert_array_almost_equal(h.corr_variations['corr__1down'], h.areas*0.8 )
    np.testing.assert_array_almost_equal(h.uncorr_variations['uncorr__1up'], h.areas*1.3 )
    np.testing.assert_array_almost_equal(h.uncorr_variations['uncorr__1down'], h.areas*0.7 )
    h = _h.slice('y', 1, name='bla')
    assert h.name == 'bla'



# Auxiliary modules
import numpy as np
from numpy.testing import assert_almost_equal as assert_approx
from os.path import join, dirname
import root_numpy as rnp
import ROOT
import sys

# Module to be tested
sys.path.append(join(dirname(__file__), '..'))
import heppy


#
# Set up
#
testfile = join(dirname(__file__), 'example.root')
hist2d_path = 'root_histogram_2d_name'
areas2d = np.array([[1., 5.],
       				[2., 6.],
       				[3., 7.],
       				[4., 8.]])
# Reference 2D ROOT histogram
reference_th2 = ROOT.TH2F('reference_th2', 'reference_th2', 4, -1.0, 1.0, 2, -1.0, 1.0)
# The zero padding is for the under-/overflows
reference_areas = np.transpose(np.array([
	[0., 0., 0., 0., 0., 0.],
	[0., 1., 2., 3., 4., 0.],
	[0., 5., 6., 7., 8., 0.],
	[0., 0., 0., 0., 0., 0.],]))
reference_area_errors = np.transpose(np.array([
	[0.0, 0.0, 0.0, 0.0, 0.0, 0.0],
	[0.0, 0.4, 0.3, 0.2, 0.1, 0.0],
	[0.0, 0.8, 0.7, 0.6, 0.5, 0.0],
	[0.0, 0.0, 0.0, 0.0, 0.0, 0.0],]))
rnp.array2hist(reference_areas, reference_th2, errors=reference_area_errors)





def test_read_TH2F():
	'''
	Reading a TH2F histogram from ROOT file
	'''
	th2 = heppy.get_TH(testfile, hist2d_path)
	assert th2.GetName() == hist2d_path
	assert th2.GetNbinsX() == 4
	assert th2.GetNbinsY() == 2
	# Remember that bin indexing in ROOT starts from 1, while 0 is the underflow bin
	assert_approx( th2.GetBinContent(4, 2), 8.0 )
	assert_approx( th2.GetBinError(3, 1), 0.2 )



def test_read_binareas_2d():
	'''
	Read bin areas of a 2D histogram from ROOT file
	'''
	areas = heppy.readroot_legacy.readroot_binareas(testfile, hist2d_path)
	np.testing.assert_array_almost_equal( areas, areas2d )



def test_read_binedges_2d():
	'''
	Read bin edges of a 2D histogram from ROOT file
	'''
	binedges = heppy.readroot_legacy.readroot_binedges2d(reference_th2)
	assert len(binedges) == 2
	np.testing.assert_array_almost_equal( binedges[0], np.array([-1.0, -0.5, 0.0, 0.5, 1.0]) )
	np.testing.assert_array_almost_equal( binedges[1], np.array([-1.0, 0.0, 1]) )



def test_read_histogram_2d():
	'''
	Test reading a ROOT TH2F and creating a Heppy histogram from it
	'''
	h = heppy.readroot_legacy.readroot(testfile, hist2d_path)
	assert isinstance(h, heppy.histogram2d)
	np.testing.assert_array_almost_equal( h.areas, areas2d )
	np.testing.assert_array_almost_equal( h.binsizes, np.ones_like(areas2d) * 0.5 )
	np.testing.assert_array_almost_equal( h.heights, areas2d / 0.5 )
	statistical__1up = (reference_areas + reference_area_errors)[1:-1,1:-1]
	np.testing.assert_array_almost_equal( h.uncorr_variations['Statistical__1up'], statistical__1up )
	statistical__1down = (reference_areas - reference_area_errors)[1:-1,1:-1]
	np.testing.assert_array_almost_equal( h.uncorr_variations['Statistical__1down'], statistical__1down )



import numpy as np
import matplotlib.pyplot as plt
import matplotlib as mpl
import heppy.panel
import heppy.histogram
from copy import deepcopy
import textwrap
import inspect

# Do not use axis value offsets by default, i.e. do not pull out common
# factors or constant offsets and put them at the top of the axis
mpl.rcParams['axes.formatter.useoffset'] = False


def _filter_plot_attributes(function, plot_attributes):
    """Returns only the plot attributes that are valid for the function."""
    updated = deepcopy(plot_attributes)
    signature = str(inspect.signature(function))
    for a in plot_attributes:
        if not ' '+a+'=' in signature:
            del updated[a]
    return updated



def _get_greater_zorder(hl):
    '''
    @param hl tuple of legend (handle, label)
    '''
    #print(hl[1])
    try:
        #print(hl[0].get_zorder())
        return(hl[0].get_zorder())
    except AttributeError:
        #print(hl[0][0].get_zorder())
        return(hl[0][0].get_zorder())



def _stable_sort_by_zorder(handles, labels):
    return zip(*sorted(zip(handles, labels), key=lambda hl: _get_greater_zorder(hl), reverse=True))



def _combine_legend_entries_with_same_label(handles, labels):
    out_handles = []
    out_labels = []
    for label in labels:
        if label in out_labels:
            # already added this label --> skip
            continue
        indices_of_label = [i for i, l in enumerate(labels) if l == label]
        # if len(indices_of_label) > 1:
        #     print(f'Found multiple artists for label {label}, will combine them into one legend entry.')
        out_labels.append((label))
        out_handles.append(tuple([handles[i] for i in reversed(indices_of_label)]))
    return out_handles, out_labels



def _get_narrowest_bin_width_in_pixels(axis, histogram):
    index_of_narrowest_bin = np.argmin(histogram.binwidths)
    x_lower = histogram.binedges[index_of_narrowest_bin]
    x_upper = histogram.binedges[index_of_narrowest_bin + 1] # cannot go out of bounds, because len(binedges) == len(binwidths) + 1
    arbitrary_y = np.max(histogram.heights[np.isfinite(histogram.heights)]) # pick some (here: maximum) finite bin height from the histogram as y coordinate
    (a, _) = axis.transData.transform((x_lower, arbitrary_y))
    (b, _) = axis.transData.transform((x_upper, arbitrary_y))
    return (b - a)



def make_figure(panels, title='', figsize=(8, 5), write='', xlims=None, xmax=None, legend_outside=False):
    '''
    :param panels: panel(s) to visualise in the plot
    :type panels: :py:class:`heppy.panel` or ``list`` of :py:class:`heppy.panel`
    :param title: plot title
    :type title: str
    :param write: may be changed to a filename, which will result in the figure being rendered and saved at the given location
    :type write: str
    :param legend_outside: option to move the legend next to the plot panels. It also changes the legend style to try to make it look better next to the plot: no box, text wrapped at 30 characters (not tested with LaTeX rendering --- proceed with caution), smaller text (fontsize='small')
    :type legend_outside: bool
    :param xlims: can be used to manually set lower and upper x-axis limits, e.g. xlims=(0.0, 2.0)
    :type xlims: `tuple` of `float`, or `None`

    :returns: tuple of the created plt.figure object and plt.axes objects. These can be assigned to variables by the user to allow further manipulations of the plot (style, contents, etc.)
    '''
    if isinstance(panels, heppy.panel):
        panels = [panels]

    if len(panels) > 1:
        fig, axes = plt.subplots(len(panels), sharex=True, gridspec_kw={'height_ratios' : [p.height for p in panels]}, figsize=figsize)
    else:
        fig, ax = plt.subplots(1, figsize=figsize)
        axes = [ax]

    artists = {}
    x = None # used to set the x-axis range later

    for i, p in enumerate(panels):

        if p.stack:
            for histogram, band in zip(p.stack.histograms, p.stack.iterbands()):
                x = band[0]
                plot_attributes = deepcopy(histogram.plot_attributes)
                if not 'label' in plot_attributes:
                    plot_attributes['label'] = histogram.name
                artists[histogram.name] = axes[i].fill_between(*band, **plot_attributes)

        for histogram in p.curves:
            x = histogram.curve()[0]
            plot_attributes = deepcopy(histogram.plot_attributes)
            if not 'label' in plot_attributes:
                plot_attributes['label'] = histogram.name
            # Filter out plot attributes that are invalid for curves
            plot_attributes = _filter_plot_attributes(mpl.lines.Line2D, plot_attributes)
            if 'label' in histogram.plot_attributes.keys():
                plot_attributes['label'] = histogram.plot_attributes['label']

            # If a histogram of the same name was already previously plotted, plot this one in the same colour.
            # Otherwise just plot it in the next colour in the sequence as usual.
            # If the histogram has an explicitly specified color, that one is always used.
            try:
                if 'color' in histogram.plot_attributes.keys():
                    raise KeyError
                previous = artists[histogram.name]
                artists[histogram.name] = axes[i].plot(*histogram.curve(), color=previous[0].get_color(), **plot_attributes)
            except (KeyError, TypeError):
                artists[histogram.name] = axes[i].plot(*histogram.curve(), **plot_attributes)

        for scatter in p.scatters:
            x = scatter.binedges[:-1]
            y = scatter.areas
            # for scatter_index in range(len(x)):
            #     print(x[scatter_index], y[scatter_index])
            plot_attributes = deepcopy(scatter.plot_attributes)
            if not 'label' in plot_attributes:
                raise RuntimeError('Please provide a label for scatter!')
            # Filter out plot attributes that are invalid for curves
            plot_attributes = _filter_plot_attributes(mpl.lines.Line2D, plot_attributes)
            if 'label' in scatter.plot_attributes.keys():
                plot_attributes['label'] = scatter.plot_attributes['label']
            # If a scatter of the same name was already previously plotted, plot this one in the same colour.
            # Otherwise just plot it in the next colour in the sequence as usual.
            # If the scatter has an explicitly specified color, that one is always used.
            try:
                if 'color' in scatter.plot_attributes.keys():
                    raise KeyError
                previous = artists[scatter.name]
                artists[scatter.name] = axes[i].plot(x, y, color=previous[0].get_color(), **plot_attributes)
            except (KeyError, TypeError):
                artists[scatter.name] = axes[i].plot(x, y, **plot_attributes)

        for histogram in p.bands:
            x = histogram.curve()[0]
            plot_attributes = deepcopy(histogram.plot_attributes)
            if not 'label' in plot_attributes:
                plot_attributes['label'] = histogram.name
            #edgecolor = '0.5' if not 'edgecolor' in plot_attributes.keys() else plot_attributes['edgecolor']
            color = plot_attributes.get('color', '0.75')
            hatch = plot_attributes.get('hatch', 'xxxxxx')
            alpha = plot_attributes.get('alpha', 0.3)
            facecolor = plot_attributes.get('color', 'none')
            artists[histogram.name] = axes[i].fill_between(*histogram.errorband(), facecolor=facecolor, edgecolor=color, alpha=alpha, linewidth=0, hatch=hatch, label=plot_attributes['label'])
            # TODO: give bands different colours and hatches

        for unbinned in p.unbinned:
            x, y, plot_attributes = unbinned
            axes[i].plot(x, y, **plot_attributes)

        pointshift = 0. - ((len(p.points) - 1) // 2) * p.pointshift # points may be shifted horizonally consecutively to prevent them from overlapping
        for histogram in p.points:
            label = histogram.plot_attributes['label'] if 'label' in histogram.plot_attributes else histogram.name
            # If a histogram of the same name was already previously plotted, plot this one in the same colour.
            # Otherwise just plot it in the next colour in the sequence as usual.
            # If the histogram has an explicitly specified color, that one is always used.
            try:
                if 'color' in histogram.plot_attributes:
                    raise KeyError
                previous = artists[histogram.name]
                #print(previous)
                #print(previous.properties())
                #color = previous[0].get_color()
                color = 'black'
            except KeyError:
                color = histogram.plot_attributes.get('color', 'k')

            marker = histogram.plot_attributes.get('marker', 'o')
            capsize = histogram.plot_attributes.get('capsize', None)

            x = histogram.curve()[0]
            errorbars = histogram.errorbars()

            # Shrink markers if bins are narrow, so that they do not overlap -->
            axes[i].set_xlim((min(x), max(x))) # need to specify axis limits first, so that we can convert minimum binwidth to axis pixels
            minwidth = _get_narrowest_bin_width_in_pixels(axes[i], histogram)
            markersize = histogram.plot_attributes.get('markersize', min(minwidth/2.3, 3.0)) # minwidth /= 2.3 needed to avoid overlapping markers
            # <-- finished picking marker size

            custom_pointshift = histogram.attributes.get('pointshift', 0.0)

            if np.count_nonzero(errorbars[0]) == 0 and np.count_nonzero(errorbars[1]) == 0:
                # Error bar sizes are all zero, do not draw error bars (this may mean that there were no variations in the histogram to begin with)
                artists[histogram.name] = axes[i].plot(*histogram.points(shift=custom_pointshift, abs_shift=True), label=label, linestyle='None', marker=marker, color=color, markersize=markersize)
            else:
                # import warnings
                # warnings.warn('Doing HACK to flip errorbars due to an inconsistency in Heppy! --- please fix this in ISSUE #23')
                no_errorbars = histogram.attributes.get('noerrorbar', False)
                if no_errorbars:
                    yerr = None
                else:
                    yerr = (errorbars[1], errorbars[0])
                xerr = None
                if histogram.attributes.get('herrorbar', False):
                    xerr = histogram.binwidths / 2.0
                artists[histogram.name] = axes[i].errorbar(*histogram.points(shift=custom_pointshift, abs_shift=True), yerr=yerr, xerr=xerr, label=label, linestyle='None', marker=marker, color=color, markersize=markersize, elinewidth=1, capsize=capsize, zorder=9999)

            pointshift += p.pointshift


        if p.title:  axes[i].set_title(p.title, x=0., y=1., va='top', ha='left', transform=axes[i].transAxes, size='large')
        if p.xlabel: axes[i].set_xlabel(p.xlabel, ha='right', x=1., size='x-large')
        if p.ylabel: axes[i].set_ylabel(p.ylabel, ha='right', y=1., size='x-large')
        if p.logx: axes[i].set_xscale('log', nonpositive='clip')
        if p.logy: axes[i].set_yscale('log', nonpositive='clip')
        if p.ylims: axes[i].set_ylim(*p.ylims)

        axes[i].tick_params(which='both', bottom=True, top=True, left=True, right=True, direction='in')
        axes[i].set_xlim((min(x), max(x)))
        if xmax:
            axes[i].set_xlim((min(x), xmax))
        if xlims:
            axes[i].set_xlim(xlims)
        legend_handles, legend_labels = axes[i].get_legend_handles_labels()
        try:
            legend_handles, legend_labels = _stable_sort_by_zorder(legend_handles, legend_labels) # do this BEFORE combining entries with same label
        except:
            pass
        legend_handles, legend_labels = _combine_legend_entries_with_same_label(legend_handles, legend_labels)
        #axes[i].legend(legend_handles[::-1], legend_labels[::-1])
        if not p.nolegend:
            # print('Updating plot legend.')
            if legend_outside:
                # Also wraps the legend labels to make the legend high and narrow --- fits well on the side of the plot
                #legend_labels = [textwrap.fill(label, width=30) for label in legend_labels]
                # Attempt to make this work better with LaTeX-rendered labels by only
                # line-breaking such labels that do not contain LaTeX math:
                legend_labels = [label if '$' in label else textwrap.fill(label, width=30) for label in legend_labels]
                axes[i].legend(legend_handles, legend_labels, bbox_to_anchor=(1.01, 1.0), fontsize='small', frameon=False, title=p.legend_title, loc=p.legend_loc)
            else:
                axes[i].legend(legend_handles, legend_labels, title=p.legend_title, loc=p.legend_loc)



    if title: axes[0].set_title(title, x=0., ha='left', size='x-large')
    fig.subplots_adjust(top=0.9, hspace=0.02)
    # fig.subplots_adjust(left=0.15, right=0.95, top=0.95, bottom=0.1, hspace=0.02)
    if legend_outside:
        fig.subplots_adjust(right=0.7)

    if write:
        fig.savefig(write)

    return fig, axes



def make_uncertainty_breakdown(histogram, separator='__', ylims=None, xlabel='', **kwargs):
    '''
    @histogram: heppy.histogram object for which the uncertainty breakdown figure will be made
    @separator: string that separates high/low (up/down, ...) indictator from the rest of the uncertainty name, e.g.
                "jet_energy_scale__1up" and "jet_energy_scale__1down" uses the separator "__"
    @ylims: may be set to a tuple/list of lower and upper y-axis limits, e.g. ylims=(0.0, 2.0)
    @**kwargs: get passed on to make_figure()
    '''
    all_variations = {**histogram.corr_variations, **histogram.uncorr_variations}
    # Make a list of unique uncertainty names that differ only by suffix (= everything after @separator), without suffixes
    uncertainty_names = sorted(list(set([key.split(separator)[0] for key in all_variations])))
    # Make a list of all the breakdown histograms
    breakdown_histograms = []
    linestyles = ['-', '--', '-.', ':']
    for i, name in enumerate(uncertainty_names):
        # There are 10 different colours in the default sequence, so change line style every 10 uncertainties to keep them distinguishable
        linestyle = linestyles[(i % 40) // 10] # index switches 0, 1, 2, 3, 0, 1, 2, 3, 0 etc. every 10 steps
        # Pick the variations corresponding to this name and make a histogram for each, but all with the same histogram name
        # (so that they will later get plotted with the same colour and get a common legend label)
        for variation in all_variations.keys():
            if not variation.startswith(name + separator):
                continue
            # Histogram whose nominal reflects the relative uncertainty in percent:
            bh = deepcopy(heppy.histdiv(histogram.extract_variation_histogram(variation), histogram, ignore_denominator_uncertainty=True))
            bh.name = name
            bh.plot_attributes = {'linestyle' : linestyle, 'label' : name}
            breakdown_histograms.append(bh)
    histogram_for_total_uncertainty = (heppy.histdiv(histogram, histogram, ignore_denominator_uncertainty=True))
    histogram_for_total_uncertainty.name = 'Total'
    histogram_for_total_uncertainty.plot_attributes = {'edgecolor' : '0.8', 'label' : 'Total'}
    line_at_100 = deepcopy(histogram_for_total_uncertainty)
    line_at_100.name = ''
    line_at_100.plot_attributes = {'color' : 'k', 'linewidth' : 1, 'linestyle' : '--'}
    # Make a panel for the figure
    panel = heppy.panel(ylabel='Ratio to nominal', curves=breakdown_histograms + [line_at_100], bands=[histogram_for_total_uncertainty], xlabel=xlabel, ylims=ylims)
    return make_figure(panel, **kwargs)



def make_roc_figure(signal, background, ylims=None, **kwargs):
    """Plots ROC curve(s) based on the given input histograms.

    :param signal: one or more histograms with "cumulative" signal efficiencies
    (= each bin contains the efficiency obtained by cutting above the lower edge
    of the bin in question)
    :type signal: :py:class:`heppy.histogram1d` or iterable thereof
    :param background: one or more histograms with "cumulative" background
    efficiencies (see `signal`). If multiple are given, they must correspond to
    the ones for signal in the same order.
    :type background: :py:class:`heppy.histogram1d` or iterable thereof
    :param **kwargs: get passed on to :py:func:`heppy.make_figure`

    Note: plot stylings and labels are taken from the _signal_ histogram(s), so
    make sure that those are set as you want them. The ones of the _background_
    histogram(s) are ignored.
    """
    rocs = []
    for sig, bkg in zip(signal, background):
        roc = deepcopy(sig)
        roc.binedges = np.append(sig.heights[::-1], np.max(sig.heights+1.0)) * 100.0 # the uppermost bin edge doesn't matter at all
        roc.areas = bkg.heights[::-1] * 100.0
        # print(roc.binedges)
        # print('SIG HEIGHTS')
        # print(roc.areas)
        # print('BKG HEIGHTS')
        # raise RuntimeError
        rocs.append(roc)
    panel = heppy.panel(scatters=rocs, ylims=ylims, xlabel='Signal efficiency (%)', ylabel='Background efficiency (%)')
    return make_figure(panel, **kwargs)
